# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Copyright 2021 Daniel Mark Gass, see __about__.py for license information.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
"""Test integer flag enumeration transform exceptions."""

import pytest
from baseline import Baseline

from plum.conformance import wrap_message
from plum.exceptions import PackError
from plum.flag import FlagX

from sample_flag import Register


register16 = FlagX(Register, byteorder="little", nbytes=2, name="register16")


class TestPack:

    """Test pack exceptions."""

    def test_not_int_val(self):
        """Test not integer values."""
        with pytest.raises(PackError) as trap:
            register16.pack("str")

        expected = Baseline(
            """
            +--------+-------+-------+------------+
            | Offset | Value | Bytes | Format     |
            +--------+-------+-------+------------+
            |        | 'str' |       | register16 |
            +--------+-------+-------+------------+

            ValueError occurred during pack operation:

            invalid literal for int() with base 10: 'str'
            """
        )

        assert wrap_message(trap.value) == expected



# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Copyright 2021 Daniel Mark Gass, see __about__.py for license information.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
"""Variable sized member definition."""

# pylint: disable=too-many-arguments

from typing import Any, Optional, Tuple

from .._typing import Format
from .._code_injector import CodeInjector
from .._default import NO_DEFAULT
from .._getbytes import getbytes
from ..dump import Record
from ..exceptions import ExcessMemoryError, InsufficientMemoryError
from .member import Member


class SizedMember(Member):

    """Sized structure member definition.

    :param doc: accessor documentation string
    :param fmt: member format, default uses type annotation
    :param size: member property which controls size
    :param ratio: number of bytes per increment of size member
    :param offset: difference in size (in bytes)
    :param default: initializer default value
    :param ignore: ignore member during comparisons
    :param readonly: block setting member attribute
    :param argrepr: format to represent member argument in structure repr

    """

    _attributes_to_copy: Tuple[str, ...] = Member._attributes_to_copy + (
        "_offset",
        "_ratio",
        "_size",
    )

    def __init__(
        self,
        doc: str,
        fmt: Format,
        size: Member,
        ratio: float,
        offset: int,
        default: Any,
        ignore: bool,
        readonly: bool,
        argrepr: Optional[str],
    ) -> None:
        super().__init__(
            doc,
            default,
            ignore,
            readonly,
            compute=False,
            argrepr=argrepr,
            fmt=fmt,
            fmt_arg=None,
        )

        if not isinstance(size, Member):
            raise TypeError("invalid 'size', must be a 'member()'")

        self._ratio = ratio
        self._offset = offset
        self._size = size

    @property
    def offset(self) -> int:
        """Difference in size (in bytes)."""
        return self._offset

    @property
    def ratio(self) -> float:
        """Number of bytes per increment of size member."""
        return self._ratio

    @property
    def size(self) -> Member:
        """Member property which controls size."""
        return self._size

    def add_name_index(
        self,
        name: str,
        index: int,
        code_injector: CodeInjector,
    ) -> int:
        """Assign name, index number, and prepare for code generation."""
        store = self.temp_store

        size_member = self.size

        index = Member.add_name_index(self, name, index, code_injector)

        if size_member.compute:
            size_member.temp_store.num_associated += 1

            # add logic to this member's setter to recompute associated size member
            # (but only when setter not explicitly defined in structure class)
            if not self.readonly and not self.fset:
                store.setter.append(
                    f"self[{size_member.index}] = None  "
                    f"# re-compute {size_member.name!r} member"
                )

            self.adjust_related_getter()

            related_pack = self.adjust_related_pack(dump=False)
            related_pack_and_dump = self.adjust_related_pack(dump=True)

            self.adjust_pack(related_pack, dump=False)
            self.adjust_pack(related_pack_and_dump, dump=True)

        self.adjust_unpack(dump=False)
        self.adjust_unpack(dump=True)

        return index

    def adjust_related_getter(self):
        """Prepare for getter code generation for related member."""
        size_member = self.size

        if not size_member.fget:
            lines = size_member.temp_store.getter
            lines[0:0] = [
                f"if self[{size_member.index}] is None:",
                f"    self[{size_member.index}] = self.unpack(self.ipack())[{size_member.index}]",
                "",
            ]

    def adjust_related_pack(self, dump: bool) -> str:
        """Prepare for pack code generation for related member."""
        size_member = self.size
        lines = (
            size_member.temp_store.pack_and_dump
            if dump
            else size_member.temp_store.pack
        )

        pack_line = lines.pop()

        dump_extra = (
            [f'    {size_member.name}_dump.value = "<skipped>"'] if dump else []
        )

        lines += (
            [
                f"{size_member.name}_pieces_index = len(pieces)",
                f"if m_{size_member.name} is None:",
            ]
            + dump_extra
            + [
                '    pieces.append(b"")',
                "else:",
                "    " + pack_line,
            ]
        )

        return pack_line

    def adjust_pack(self, related_pack_line: str, dump: bool):
        """Prepare for pack code generation for this member."""
        size_member = self.size
        lines = self.temp_store.pack_and_dump if dump else self.temp_store.pack

        size = f'len(b"".join(pieces[{self.name}_pieces_index:]))'

        if self.ratio != 1:
            size = f"int({size} // {self.ratio})"

        if self.offset != 0:
            size += f" + {self.offset}"

        pack_line = lines.pop()

        # pylint: disable=f-string-without-interpolation
        lines += [
            f"if m_{size_member.name} is None:",
            f"    {self.name}_pieces_index = len(pieces)",
            f"    " + pack_line,
            f"    m_{size_member.name} = {size}",
            f"    " + related_pack_line,
            f"    pieces[{size_member.name}_pieces_index] = pieces.pop()",
            f"else:",
            f"    " + pack_line,
        ]

    def adjust_unpack(self, dump: bool) -> None:
        """Prepare for unpack code generation for this member."""
        size_member = self.size
        store = self.temp_store
        lines = store.unpack_and_dump if dump else store.unpack

        temp_buffer = f"{self.name}_buffer"

        unpack_line = lines.pop()
        dump_arg = f", {self.name}_dump" if dump else ""

        nbytes = f"m_{size_member.name}"
        if self.offset:
            nbytes = f"({nbytes} - {self.offset})"
        if self.ratio != 1:
            nbytes = f"int({nbytes} * {self.ratio})"

        if self.offset or self.ratio != 1:
            lines.append(f"{self.name}_nbytes = {nbytes}")
            nbytes = f"{self.name}_nbytes"

        lines += [
            f"buffer, {temp_buffer} = buffer[:offset + {nbytes}], buffer",
            f"if len(buffer) < offset + {nbytes}:",
            f"    cls.{self.name}.report_insufficient_bytes(buffer, offset, {nbytes}{dump_arg})",
            unpack_line,
            "if offset < len(buffer):",
            f"    cls.{self.name}.report_extra_bytes(buffer[offset:]{dump_arg})",
            f"buffer = {temp_buffer}",
        ]

    @staticmethod
    def report_insufficient_bytes(
        buffer: bytes, offset: int, nbytes: int, dump: Optional[Record] = None
    ):
        """Add unconsumed bytes leftover from unpacking to dump summary."""
        if dump is None:
            raise InsufficientMemoryError(
                "insufficient bytes, redo-ing to include dump"
            )

        getbytes(buffer, offset, dump, nbytes)

    @staticmethod
    def report_extra_bytes(extra_bytes: bytes, dump: Optional[Record] = None) -> None:
        """Add unconsumed bytes leftover from unpacking to dump summary."""
        if dump is not None:
            for i in range(0, len(extra_bytes), 16):
                if i:
                    dump.add_record(memory=extra_bytes[i : i + 16])
                else:
                    dump.add_record(
                        separate=True,
                        value="<excess bytes>",
                        memory=extra_bytes[i : i + 16],
                    )

        raise ExcessMemoryError(extra_bytes)


def sized_member(
    doc: str = "",
    *,
    fmt: Format,
    size: Any,  # must be Member, checked at runtime
    ratio: float = 1,
    offset: int = 0,
    default: Any = NO_DEFAULT,
    ignore: bool = False,
    readonly: bool = False,
    argrepr: Optional[str] = None,
) -> Any:
    """Sized structure member definition.

    :param doc: accessor documentation string
    :param fmt: member format, default uses type annotation
    :param size: member property which controls size
    :param int ratio: number of bytes per increment of size member
    :param int offset: difference in size (in bytes)
    :param default: initializer default value
    :param ignore: ignore member during comparisons
    :param readonly: block setting member attribute
    :param argrepr: format to represent member argument in structure repr

    """
    return SizedMember(
        doc, fmt, size, ratio, offset, default, ignore, readonly, argrepr
    )



# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Copyright 2021 Daniel Mark Gass, see __about__.py for license information.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
"""Structure member definition."""

# pylint: disable=too-many-instance-attributes,too-many-public-methods,too-many-arguments

from dataclasses import dataclass, field
from types import FunctionType, MethodType, BuiltinFunctionType, BuiltinMethodType
from typing import Any, Generator, List, Optional, Tuple, Type, Union

from .._code_injector import CodeInjector
from .._default import NO_DEFAULT
from .._typing import FactoryFormat
from ..enum import EnumX
from ..flag import FlagX
from ..data import Data, DataMeta
from ..transform import Transform

FUNCTION_TYPES = (FunctionType, MethodType, BuiltinFunctionType, BuiltinMethodType)


@dataclass
class TempStore:

    """Temporary information store for generating implementation code."""

    num_associated: int = 0

    local_name: str = ""
    cls_name: str = ""
    type_hint: str = ""

    init_parameter: str = ""
    init_pre_begin: List[str] = field(default_factory=list)
    init_begin: List[str] = field(default_factory=list)
    init_end: List[str] = field(default_factory=list)

    pack: List[str] = field(default_factory=list)
    pack_and_dump: List[str] = field(default_factory=list)

    unpack: List[str] = field(default_factory=list)
    unpack_and_dump: List[str] = field(default_factory=list)
    unpack_checks: List[str] = field(default_factory=list)

    getter: List[str] = field(default_factory=list)
    setter: List[str] = field(default_factory=list)


class Member(property):

    """Structure member definition base class.

    :param doc: accessor documentation string
    :param default: initializer default value
    :param ignore: ignore member during comparisons
    :param readonly: block setting member attribute
    :param compute: initializer defaults to compute based on another member
    :param argrepr: format to represent member argument in structure repr
    :param fmt: member format
    :param fmt_arg: member property to use as format factory argument

    """

    _nbytes: Optional[int]
    _index: Optional[int]
    _temp_store: Optional[TempStore]

    _attributes_to_copy: Tuple[str, ...] = (
        "_argrepr",
        "_compute",
        "_doc",
        "_default",
        "_fmt",
        "_fmt_arg",
        "_ignore",
        "_nbytes",
        "_readonly",
    )

    def __init__(
        self,
        doc: str,
        default: Any,
        ignore: bool,
        readonly: bool,
        compute: bool,
        argrepr: Optional[str],
        fmt: FactoryFormat,
        fmt_arg: Optional["Member"],
    ) -> None:
        property.__init__(self)

        if compute and default is not NO_DEFAULT:
            raise TypeError("'default' may not be specified when 'compute=True'")

        if isinstance(default, property):
            default = default.fget

        nbytes = None

        if isinstance(fmt, (DataMeta, Transform)):
            if fmt_arg is not None:
                raise TypeError(
                    "when 'fmt_arg' specified, 'fmt' must be a factory function"
                )
            nbytes = fmt.__nbytes__

        else:
            if isinstance(fmt, property):
                fmt = fmt.fget

            if isinstance(fmt, FUNCTION_TYPES):
                if not (isinstance(fmt_arg, Member) or fmt_arg is None):
                    raise TypeError("'fmt_arg' must be a structure 'member()'")
            else:
                raise TypeError(
                    "'fmt' must be a data store, transform, or a factory function"
                )

        self._nbytes = nbytes
        self._doc = doc
        self._default = default
        self._ignore = ignore
        self._readonly = readonly
        self._compute = compute
        self._argrepr = argrepr
        self._fmt = fmt
        self._fmt_arg = fmt_arg

        self._name = ""  # assigned during structure class construction
        self._index = None  # assigned during structure class construction

        self._temp_store = None

    @property
    def temp_store(self) -> TempStore:
        """Temporary information store for generating implementation code."""
        store = self._temp_store

        if store is None:
            self._temp_store = store = TempStore()

        return store

    @temp_store.deleter
    def temp_store(self):
        self._temp_store = None

    @property
    def argrepr(self) -> str:
        """Format string to represent member argument in structure repr."""
        if isinstance(self._argrepr, str):
            repr_format = self._argrepr

        elif isinstance(self._fmt, (EnumX, FlagX)):
            repr_format = (
                f"{self.name}={{repr(self.{self.name}).split(':', 1)[0].lstrip('<')}}"
            )

        else:
            repr_format = f"{self.name}={{self.{self.name}!r}}"

        return repr_format

    @property
    def compute(self) -> bool:
        """Initializer defaults to compute based on another member."""
        return self._compute

    @property
    def default_is_a_factory(self) -> bool:
        """Member default is produced by a factory function."""
        return self.compute or isinstance(self.default, FUNCTION_TYPES)

    @property
    def fmt_is_a_factory(self) -> bool:
        """Member format is produced by a factory function."""
        return isinstance(self.fmt, FUNCTION_TYPES)

    @property
    def default(self) -> Any:
        """Structure initializer default value for this structure member."""
        return self._default

    @property
    def doc(self) -> str:
        """Member description."""
        return self._doc

    @property
    def fmt(self) -> Union[Transform, Type[Data]]:
        """Member format."""
        return self._fmt

    @property
    def fmt_arg(self) -> Optional["Member"]:
        """Member property to use as format factory argument."""
        return self._fmt_arg

    @property
    def ignore(self) -> bool:
        """Ignore member when compared against structure of same type."""
        return self._ignore

    @property
    def index(self) -> int:
        """Member number."""
        if self._index is None:  # pragma: no cover (for mypy)
            raise RuntimeError(
                "internal error, likely cause was getter/setter not immediately "
                "following member"
            )
        return self._index

    @index.setter
    def index(self, value: int):
        self._index = value

    @property
    def name(self) -> str:
        """Bit field name."""
        if self._name is None:  # pragma: no cover (for mypy)
            raise RuntimeError("internal error")
        return self._name

    @property
    def nbytes(self) -> Optional[int]:
        """Transform format size in bytes."""
        return self._nbytes

    @property
    def readonly(self) -> bool:
        """Block setting member attribute."""
        return self._readonly

    def copy(self, fget=None, fset=None, name="", index=None):
        """Copy structure member property definition.

        Copy every member property attribute except override with values
        provided as arguments (except always dump temp_store).

        """
        cls = type(self)

        copy = cls.__new__(cls)

        property.__init__(copy, fget, fset, self.fdel)

        for attr_name in self._attributes_to_copy:
            setattr(copy, attr_name, getattr(self, attr_name))

        # pylint: disable=protected-access
        copy._name = name
        copy._index = index
        copy._temp_store = None

        self._temp_store = None

        return copy

    def getter(self, fget):
        """Decorator to change getter method."""
        return self.copy(fget, self.fset, self._name, self._index)

    def setter(self, fset):
        """Decorator to change setter method."""
        if self.readonly:
            raise TypeError(
                "'setter' not allowed on read-only structure member properties"
            )

        return self.copy(self.fget, fset, self._name, self._index)

    def deleter(self, fdel):
        """Decorator to change deleter method."""
        raise TypeError("structure member properties do not support 'deleter'")

    def add_name_index(
        self,
        name: str,
        index: int,
        code_injector: CodeInjector,
    ) -> int:
        """Assign name, index number, and prepare for code generation."""
        # pylint: disable=too-many-branches,too-many-statements,unused-argument
        if self._name:
            raise TypeError(
                f"invalid structure member {name!r} definition, "
                f"member instance can not be shared, create a new instance"
            ) from None

        self._name = name
        self.index = index

        store = self.temp_store

        if self.fmt_is_a_factory:
            store.cls_name = f"{name}_fmt"
        else:
            store.cls_name = f"cls.{name}.fmt"

        store.local_name = "m_" + name
        store.type_hint = code_injector.get_type_hint_expression(name)

        store.init_parameter = self.get_init_parameter()

        if self.default is not NO_DEFAULT and not self.compute:
            init_expr = f"type(self).{name}.default"
            if self.default_is_a_factory:
                target_store = store.init_end
                access_expr = f"self[{self.index}]"
                init_expr += "(self)"
            else:
                target_store = store.init_pre_begin
                access_expr = name
                init_expr = code_injector.get_expression(self.default, init_expr)

            target_store.extend(
                [
                    f"if {name} is None:",
                    f"    {access_expr} = {init_expr}",
                ]
            )

        if not self.fget:
            store.getter.append(f"return self[{self.index}]")

        if not self.readonly and not self.fset:
            store.setter.append(f"self[{self.index}] = value")

        dump = f'dump.add_record(access="{name}", fmt={store.cls_name})'
        extra = self.extra_args

        if self.fmt_is_a_factory:
            if self.fmt_arg is None:
                cls_factory = [f"{store.cls_name} = cls.{name}.fmt(value)"]
            else:
                cls_factory = [
                    f"{store.cls_name} = cls.{name}.fmt(m_{self.fmt_arg.name})"
                ]
        else:
            cls_factory = []

        store.pack += cls_factory + [
            f"{store.cls_name}.__pack__(m_{name}, pieces, dump{extra})",
        ]
        store.pack_and_dump += cls_factory + [
            f"{name}_dump = {dump}",
            f"{store.cls_name}.__pack__(m_{name}, pieces, {name}_dump{extra})",
        ]

        if self.fmt_is_a_factory:
            if self.fmt_arg is None:
                # FUTURE change structure -> value
                cls_factory = [f"{store.cls_name} = cls.{name}.fmt(structure)"]
            else:
                cls_factory = [
                    f"{store.cls_name} = cls.{name}.fmt(m_{self.fmt_arg.name})"
                ]
        else:
            cls_factory = []

        store.unpack += cls_factory + [
            f"m_{name}, offset = {store.cls_name}.__unpack__(buffer, offset, dump{extra})",
        ]

        store.unpack_and_dump += cls_factory + [
            f"{name}_dump = {dump}",
            f"m_{name}, offset = {store.cls_name}.__unpack__(buffer, offset, {name}_dump{extra})",
        ]

        if (
            self.readonly
            and (self.default is not NO_DEFAULT)
            and not self.default_is_a_factory
        ):
            alternative = f"cls.{name}.default"
            default = code_injector.get_expression(self.default, alternative)
            store.unpack_checks += [
                f"if m_{name} != {default}:",
                f'    raise ValueError(f"{name!r} must be {{{default}}}")',
            ]

        return index + 1

    def iter_getter_lines(self) -> Generator[str, None, None]:
        """Iterate generated getter implementation lines."""

        store = self.temp_store

        lines = store.getter

        if lines:
            type_hint = f" -> {store.type_hint}" if store.type_hint else ""

            yield f"@{self.name}.getter"
            yield f"def {self.name}(self){type_hint}:"
            if self.doc:
                yield f'    """{self.doc}"""'
            for line in lines:
                yield ("    " + line).rstrip()
            yield ""

    def iter_setter_lines(self) -> Generator[str, None, None]:
        """Iterate generated setter implementation lines."""
        store = self.temp_store

        lines = store.setter

        if lines:
            type_hint = f": {store.type_hint}" if store.type_hint else ""

            yield f"@{self.name}.setter"
            yield f"def {self.name}(self, value{type_hint}) -> None:"
            if self.doc:
                yield f'    """{self.doc}"""'
            for line in lines:
                yield ("    " + line).rstrip()
            yield ""

    def get_init_parameter(self) -> str:
        """Get member parameter for structure __init__ method.

        Include type annotation where possible.

        """
        parameter = self.name

        type_hint = self.temp_store.type_hint
        is_optional = self.default is not NO_DEFAULT or self.compute

        if type_hint:
            if is_optional:
                type_hint = f"Optional[{type_hint}]"

            parameter += f": {type_hint}"

        if is_optional:
            parameter += " = None"

        return parameter

    @property
    def extra_args(self) -> str:
        """Additional __pack__ and __unpack__ arguments (for code generation)."""
        return ""

    def check(self):
        """Perform final checks."""
        if self.compute and not self.temp_store.num_associated:
            raise TypeError(
                f"{self.name!r} member never associated with member "
                f"used to compute it"
            )

        # self._temp_store = None

    def __repr__(self):
        return f"{type(self).__name__}(name={self.name!r})"


def member(
    doc: str = "",
    *,
    default: Any = NO_DEFAULT,
    ignore: bool = False,
    readonly: bool = False,
    compute: bool = False,
    argrepr: Optional[str] = None,
    fmt: FactoryFormat,
    fmt_arg: Optional["Member"] = None,
) -> Any:
    """Define structure member properties.

    :param doc: accessor documentation string
    :param default: initializer default value
    :param ignore: ignore member during comparisons
    :param readonly: block setting member attribute
    :param compute: initializer defaults to compute based on another member
    :param argrepr: format to represent member argument in structure repr
    :param fmt: member format
    :param fmt_arg: member property to use as format factory argument

    """
    return Member(doc, default, ignore, readonly, compute, argrepr, fmt, fmt_arg)



# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Copyright 2021 Daniel Mark Gass, see __about__.py for license information.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
"""Test Structure compare."""

# pylint: disable=unneeded-not

import pytest
from baseline import Baseline

from plum.bigendian import uint8
from plum.structure import Structure, member, sized_member


class Base:

    """Test case base class."""

    Struct: Structure
    implementation_baseline: Baseline

    @staticmethod
    def iter_lines(code):
        lines = code.split("\n")

        active = False

        for line in lines:
            if "__eq__" in line or "__ne__" in line:
                active = True

            if active:
                yield line

            if not line.strip():
                active = False

    def test_generated_code(self):
        compare_implementation = "\n".join(self.iter_lines(self.Struct.implementation))
        assert compare_implementation.rstrip() == self.implementation_baseline


class TestVanilla(Base):

    """Test structure compare when no compute or ignore."""

    class Struct(Structure):

        m0: int = member(fmt=uint8)
        m1: int = member(fmt=uint8)

    implementation_baseline = Baseline(
        """
        def __eq__(self, other: Any) -> bool:
            if isinstance(other, dict):
                other = self._make_structure_from_dict(other)
            return list.__eq__(self, other)

        def __ne__(self, other: Any) -> bool:
            if isinstance(other, dict):
                other = self._make_structure_from_dict(other)
            return list.__ne__(self, other)
        """
    )

    struct1 = Struct(m0=0, m1=1)
    struct2 = Struct(m0=0, m1=2)

    def test_eq(self):
        assert self.struct1 == self.struct1
        assert not self.struct1 == self.struct2

    def test_ne(self):
        assert self.struct1 != self.struct2
        assert not self.struct1 != self.struct1


class TestIgnoredMember(Base):

    """Test structure compare when member set to ignore."""

    class Struct(Structure):

        m0: int = member(fmt=uint8)
        m1: int = member(fmt=uint8, ignore=True)

    implementation_baseline = Baseline(
        """
        def __eq__(self, other: Any) -> bool:
            if isinstance(other, dict):
                other = self._make_structure_from_dict(other)
            elif isinstance(other, type(self)):
                _s_m0, _s_m1 = self
                _o_m0, _o_m1 = other
                return _s_m0 == _o_m0
            else:
                return list.__eq__(self, other)

        def __ne__(self, other: Any) -> bool:
            if isinstance(other, dict):
                other = self._make_structure_from_dict(other)
            elif isinstance(other, type(self)):
                _s_m0, _s_m1 = self
                _o_m0, _o_m1 = other
                return _s_m0 != _o_m0
            else:
                return list.__ne__(self, other)
        """
    )

    struct1 = Struct(m0=0, m1=1)
    struct2 = Struct(m0=0, m1=2)
    struct3 = Struct(m0=3, m1=2)

    def test_eq(self):
        assert self.struct1 == self.struct2
        assert not self.struct1 == self.struct3

    def test_ne(self):
        assert self.struct1 != self.struct3
        assert not self.struct1 != self.struct2


class TestOneCompute(Base):

    """Test structure compare when one member computed."""

    class Struct(Structure):

        size: int = member(fmt=uint8, compute=True)
        sized: int = sized_member(fmt=uint8, size=size)

    implementation_baseline = Baseline(
        """
        def __eq__(self, other: Any) -> bool:
            if isinstance(other, dict):
                other = self._make_structure_from_dict(other)
            elif isinstance(other, type(self)):
                _s_size, _s_sized = self
                if _s_size is None:
                    _s_size, _s_sized = self.unpack(self.ipack())
                _o_size, _o_sized = other
                if _o_size is None:
                    _o_size, _o_sized = self.unpack(other.ipack())
                return (_s_size, _s_sized) == (_o_size, _o_sized)
            else:
                return list.__eq__(self, other)

        def __ne__(self, other: Any) -> bool:
            if isinstance(other, dict):
                other = self._make_structure_from_dict(other)
            elif isinstance(other, type(self)):
                _s_size, _s_sized = self
                if _s_size is None:
                    _s_size, _s_sized = self.unpack(self.ipack())
                _o_size, _o_sized = other
                if _o_size is None:
                    _o_size, _o_sized = self.unpack(other.ipack())
                return (_s_size, _s_sized) != (_o_size, _o_sized)
            else:
                return list.__ne__(self, other)
        """
    )

    struct1 = Struct(sized=2)
    struct2 = Struct(size=1, sized=2)
    struct3 = Struct(size=3, sized=2)
    struct4 = Struct(sized=4)

    def test_eq(self):
        assert self.struct1 == self.struct1
        assert self.struct1 == self.struct2
        assert not self.struct1 == self.struct3
        assert not self.struct1 == self.struct4

    def test_ne(self):
        assert self.struct1 != self.struct3
        assert self.struct1 != self.struct4
        assert not self.struct1 != self.struct1
        assert not self.struct1 != self.struct2


class TestTwoComputes(Base):

    """Test structure compare when one member computed."""

    class Struct(Structure):

        size: int = member(fmt=uint8, compute=True)
        sized: int = sized_member(fmt=uint8, size=size)

        size2: int = member(fmt=uint8, compute=True)
        sized2: int = sized_member(fmt=uint8, size=size2)

    implementation_baseline = Baseline(
        """
        def __eq__(self, other: Any) -> bool:
            if isinstance(other, dict):
                other = self._make_structure_from_dict(other)
            elif isinstance(other, type(self)):
                _s_size, _s_sized, _s_size2, _s_sized2 = self
                if None in (_s_size, _s_size2):
                    _s_size, _s_sized, _s_size2, _s_sized2 = self.unpack(self.ipack())
                _o_size, _o_sized, _o_size2, _o_sized2 = other
                if None in (_o_size, _o_size2):
                    _o_size, _o_sized, _o_size2, _o_sized2 = self.unpack(other.ipack())
                return (_s_size, _s_sized, _s_size2, _s_sized2) == (_o_size, _o_sized, _o_size2, _o_sized2)
            else:
                return list.__eq__(self, other)

        def __ne__(self, other: Any) -> bool:
            if isinstance(other, dict):
                other = self._make_structure_from_dict(other)
            elif isinstance(other, type(self)):
                _s_size, _s_sized, _s_size2, _s_sized2 = self
                if None in (_s_size, _s_size2):
                    _s_size, _s_sized, _s_size2, _s_sized2 = self.unpack(self.ipack())
                _o_size, _o_sized, _o_size2, _o_sized2 = other
                if None in (_o_size, _o_size2):
                    _o_size, _o_sized, _o_size2, _o_sized2 = self.unpack(other.ipack())
                return (_s_size, _s_sized, _s_size2, _s_sized2) != (_o_size, _o_sized, _o_size2, _o_sized2)
            else:
                return list.__ne__(self, other)
        """
    )

    struct1 = Struct(sized=2, sized2=7)
    struct2 = Struct(size=1, sized=2, sized2=7)
    struct3 = Struct(size=3, sized=2, size2=1, sized2=7)
    struct4 = Struct(sized=4, sized2=7)

    def test_eq(self):
        assert self.struct1 == self.struct1
        assert self.struct1 == self.struct2
        assert not self.struct1 == self.struct3
        assert not self.struct1 == self.struct4

    def test_ne(self):
        assert self.struct1 != self.struct3
        assert self.struct1 != self.struct4
        assert not self.struct1 != self.struct1
        assert not self.struct1 != self.struct2


class TestIgnoredCompute(Base):

    """Test structure compare when one member computed."""

    class Struct(Structure):

        size: int = member(fmt=uint8, compute=True, ignore=True)
        sized: int = sized_member(fmt=uint8, size=size)

    implementation_baseline = Baseline(
        """
        def __eq__(self, other: Any) -> bool:
            if isinstance(other, dict):
                other = self._make_structure_from_dict(other)
            elif isinstance(other, type(self)):
                _s_size, _s_sized = self
                _o_size, _o_sized = other
                return _s_sized == _o_sized
            else:
                return list.__eq__(self, other)

        def __ne__(self, other: Any) -> bool:
            if isinstance(other, dict):
                other = self._make_structure_from_dict(other)
            elif isinstance(other, type(self)):
                _s_size, _s_sized = self
                _o_size, _o_sized = other
                return _s_sized != _o_sized
            else:
                return list.__ne__(self, other)
        """
    )

    struct1 = Struct(sized=2)
    struct2 = Struct(size=1, sized=2)
    struct3 = Struct(size=3, sized=2)
    struct4 = Struct(sized=4)
    struct5 = Struct(size=1, sized=4)

    def test_eq(self):
        assert self.struct1 == self.struct1
        assert self.struct1 == self.struct2
        assert self.struct1 == self.struct3
        assert not self.struct1 == self.struct4
        assert not self.struct1 == self.struct5

    def test_ne(self):
        assert self.struct1 != self.struct4
        assert self.struct1 != self.struct5
        assert not self.struct1 != self.struct1
        assert not self.struct1 != self.struct2
        assert not self.struct1 != self.struct3


class TestIgnoreEverything(Base):

    """Test that define eq method is not overridden."""

    class Struct(Structure):

        m0: int = member(fmt=uint8, ignore=True)

    implementation_baseline = Baseline(
        """
        def __eq__(self, other: Any) -> bool:
            if isinstance(other, dict):
                other = self._make_structure_from_dict(other)
            elif isinstance(other, type(self)):
                return True
            else:
                return list.__eq__(self, other)

        def __ne__(self, other: Any) -> bool:
            if isinstance(other, dict):
                other = self._make_structure_from_dict(other)
            elif isinstance(other, type(self)):
                return False
            else:
                return list.__ne__(self, other)
        """
    )

    struct1 = Struct(m0=1)
    struct2 = Struct(m0=2)

    def test_eq(self):
        assert self.struct1 == self.struct2
        assert self.struct1 == [1]

    def test_ne(self):
        assert not self.struct1 != self.struct2
        assert not self.struct1 != [1]


class TestOverrideEq(Base):

    """Test that custom __eq__ method is not overridden."""

    class Struct(Structure):

        m0: int = member(fmt=uint8)

        def __eq__(self, other):
            raise RuntimeError("hit eq")

    implementation_baseline = Baseline(
        """
        def __ne__(self, other: Any) -> bool:
            if isinstance(other, dict):
                other = self._make_structure_from_dict(other)
            return list.__ne__(self, other)
        """
    )

    def test_eq(self):
        with pytest.raises(RuntimeError) as trap:
            assert self.Struct(m0=1) == self.Struct(m0=1)  # pylint: disable=no-member

        assert str(trap.value) == "hit eq"


class TestOverrideNe(Base):

    """Test that custom __ne__ method is not overridden."""

    class Struct(Structure):

        m0: int = member(fmt=uint8)

        def __ne__(self, other):
            raise RuntimeError("hit ne")

    implementation_baseline = Baseline(
        """
        def __eq__(self, other: Any) -> bool:
            if isinstance(other, dict):
                other = self._make_structure_from_dict(other)
            return list.__eq__(self, other)
        """
    )

    def test_ne(self):
        with pytest.raises(RuntimeError) as trap:
            assert self.Struct(m0=1) != self.Struct(m0=1)  # pylint: disable=no-member

        assert str(trap.value) == "hit ne"


class TestDict:

    """Test structure compare when no compute or ignore."""

    class Struct(Structure):

        m0: int = member(fmt=uint8)
        m1: int = member(fmt=uint8)

    struct = Struct(m0=0, m1=1)

    def test_eq(self):
        assert self.struct == dict(m0=0, m1=1)
        assert not self.struct == dict(m0=0, m1=2)

    def test_ne(self):
        assert self.struct != dict(m0=0, m1=2)
        assert not self.struct != dict(m0=0, m1=1)



# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Copyright 2021 Daniel Mark Gass, see __about__.py for license information.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
"""Test items transform properties."""

import pytest

from plum.exceptions import SizeError
from plum.items import ItemsX
from plum.littleendian import uint8


class TestDefault:

    """Test with as many left to default as possible."""

    itemsx = ItemsX()

    def test_name(self):
        assert self.itemsx.name == "Any"

    def test_hint(self):
        assert self.itemsx.name == "Any"

    def test_nbytes(self):
        with pytest.raises(SizeError):
            return self.itemsx.nbytes

    def test_fmt(self):
        assert self.itemsx.fmt is None


class TestPositional:

    """Test explicitly defined with positional argument."""

    itemsx = ItemsX(uint8, "name")

    def test_hint(self):
        assert self.itemsx.__hint__ == "int"

    def test_name(self):
        assert self.itemsx.name == "name"

    def test_nbytes(self):
        assert self.itemsx.nbytes == 1

    def test_fmt(self):
        assert self.itemsx.fmt is uint8


class TestKeyword(TestPositional):

    """Test explicitly defined with keyword argument."""

    itemsx = ItemsX(fmt=uint8, name="name")


class TestDefaultNameHint:
    def test_none(self):
        xform = ItemsX(fmt=None)
        assert xform.name == xform.__hint__ == "Any"

    def test_tuple(self):
        xform = ItemsX(fmt=(None, uint8))
        assert xform.name == xform.__hint__ == "Tuple[Any, int]"

    def test_list_uniform(self):
        xform = ItemsX(fmt=[uint8, uint8])
        assert xform.name == xform.__hint__ == "List[int]"

    def test_list_varied(self):
        xform = ItemsX(fmt=[uint8, None])
        assert xform.name == xform.__hint__ == "List[Any]"

    def test_dict_uniform(self):
        xform = ItemsX(fmt={"a": uint8, "b": uint8})
        assert xform.name == xform.__hint__ == "Dict[str, int]"

    def test_dict_varied(self):
        xform = ItemsX(fmt={"a": uint8, "b": None})
        assert xform.name == xform.__hint__ == "Dict[str, Any]"

    def test_nested(self):
        xform = ItemsX(fmt=((None, uint8), {"a": uint8}))
        assert xform.name == xform.__hint__ == "Tuple[Tuple[Any, int], Dict[str, int]]"



# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Copyright 2021 Daniel Mark Gass, see __about__.py for license information.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
"""Test format factory function as format."""

import pytest
from baseline import Baseline

from plum.conformance import Case, CaseData, wrap_message
from plum.littleendian import uint8, uint16
from plum.structure import Structure, member


class TestFmtArg(Case):

    """Factory passed member as controlled by 'fmt_arg'."""

    class Struct(Structure):

        datatype: int = member(fmt=uint8)
        data: int = member(fmt_arg=datatype, fmt={0: uint8, 1: uint16}.__getitem__)
        bookend: int = member(fmt=uint8)

    data = CaseData(
        fmt=Struct,
        bindata=bytes.fromhex("000177"),
        nbytes=None,
        values=(
            Struct(datatype=0, data=1, bookend=0x77),
            [0, 1, 0x77],
        ),
        dump=Baseline(
            """
            +--------+----------+-------+-------+--------------------+
            | Offset | Access   | Value | Bytes | Format             |
            +--------+----------+-------+-------+--------------------+
            |        |          |       |       | Struct (Structure) |
            | 0      | datatype | 0     | 00    | uint8              |
            | 1      | data     | 1     | 01    | uint8              |
            | 2      | bookend  | 119   | 77    | uint8              |
            +--------+----------+-------+-------+--------------------+
            """
        ),
        excess=Baseline(
            """
            +--------+----------+----------------+-------+--------------------+
            | Offset | Access   | Value          | Bytes | Format             |
            +--------+----------+----------------+-------+--------------------+
            |        |          |                |       | Struct (Structure) |
            | 0      | datatype | 0              | 00    | uint8              |
            | 1      | data     | 1              | 01    | uint8              |
            | 2      | bookend  | 119            | 77    | uint8              |
            +--------+----------+----------------+-------+--------------------+
            | 3      |          | <excess bytes> | 99    |                    |
            +--------+----------+----------------+-------+--------------------+

            ExcessMemoryError occurred during unpack operation:

            1 unconsumed bytes
            """
        ),
        shortage=Baseline(
            """
            +--------+----------+----------------------+-------+--------------------+
            | Offset | Access   | Value                | Bytes | Format             |
            +--------+----------+----------------------+-------+--------------------+
            |        |          |                      |       | Struct (Structure) |
            | 0      | datatype | 0                    | 00    | uint8              |
            | 1      | data     | 1                    | 01    | uint8              |
            |        | bookend  | <insufficient bytes> |       | uint8              |
            +--------+----------+----------------------+-------+--------------------+

            InsufficientMemoryError occurred during unpack operation:

            1 too few bytes to unpack uint8, 1 needed, only 0 available
            """
        ),
        implementation=Baseline(
            """
            def __init__(self, *, datatype: int, data: int, bookend: int) -> None:
                self[:] = (datatype, data, bookend)

            @datatype.getter
            def datatype(self) -> int:
                return self[0]

            @datatype.setter
            def datatype(self, value: int) -> None:
                self[0] = value

            @data.getter
            def data(self) -> int:
                return self[1]

            @data.setter
            def data(self, value: int) -> None:
                self[1] = value

            @bookend.getter
            def bookend(self) -> int:
                return self[2]

            @bookend.setter
            def bookend(self, value: int) -> None:
                self[2] = value

            @classmethod
            def __pack__(cls, value, pieces: List[bytes], dump: Optional[Record] = None) -> None:
                if isinstance(value, dict):
                    value = cls._make_structure_from_dict(value)

                (m_datatype, m_data, m_bookend) = value

                if dump is None:
                    cls.datatype.fmt.__pack__(m_datatype, pieces, dump)

                    data_fmt = cls.data.fmt(m_datatype)
                    data_fmt.__pack__(m_data, pieces, dump)

                    cls.bookend.fmt.__pack__(m_bookend, pieces, dump)

                else:
                    datatype_dump = dump.add_record(access="datatype", fmt=cls.datatype.fmt)
                    cls.datatype.fmt.__pack__(m_datatype, pieces, datatype_dump)

                    data_fmt = cls.data.fmt(m_datatype)
                    data_dump = dump.add_record(access="data", fmt=data_fmt)
                    data_fmt.__pack__(m_data, pieces, data_dump)

                    bookend_dump = dump.add_record(access="bookend", fmt=cls.bookend.fmt)
                    cls.bookend.fmt.__pack__(m_bookend, pieces, bookend_dump)

            @classmethod
            def __unpack__(cls, buffer: bytes, offset: int, dump: Optional[Record] = None) -> Tuple["Struct", int]:
                structure = list.__new__(cls)

                if dump is None:
                    m_datatype, offset = cls.datatype.fmt.__unpack__(buffer, offset, dump)

                    data_fmt = cls.data.fmt(m_datatype)
                    m_data, offset = data_fmt.__unpack__(buffer, offset, dump)

                    m_bookend, offset = cls.bookend.fmt.__unpack__(buffer, offset, dump)

                else:
                    datatype_dump = dump.add_record(access="datatype", fmt=cls.datatype.fmt)
                    m_datatype, offset = cls.datatype.fmt.__unpack__(buffer, offset, datatype_dump)

                    data_fmt = cls.data.fmt(m_datatype)
                    data_dump = dump.add_record(access="data", fmt=data_fmt)
                    m_data, offset = data_fmt.__unpack__(buffer, offset, data_dump)

                    bookend_dump = dump.add_record(access="bookend", fmt=cls.bookend.fmt)
                    m_bookend, offset = cls.bookend.fmt.__unpack__(buffer, offset, bookend_dump)

                structure[:] = (m_datatype, m_data, m_bookend)

                return structure, offset

            def __eq__(self, other: Any) -> bool:
                if isinstance(other, dict):
                    other = self._make_structure_from_dict(other)
                return list.__eq__(self, other)

            def __ne__(self, other: Any) -> bool:
                if isinstance(other, dict):
                    other = self._make_structure_from_dict(other)
                return list.__ne__(self, other)

            def __repr__(self) -> str:
                try:
                    return f"{type(self).__name__}(datatype={self.datatype!r}, data={self.data!r}, bookend={self.bookend!r})"
                except Exception:
                    return f"{type(self).__name__}()"

            """
        ),
    )


class TestMethod(Case):

    """Factory passed half baked structure instance (factory is a simple method)."""

    class Struct(Structure):
        def datatype_fmt(self):
            mapping = {0: uint8, 1: uint16}
            return mapping[self[0]]

        datatype: int = member(fmt=uint8)
        data: int = member(fmt=datatype_fmt)
        bookend: int = member(fmt=uint8)

    data = CaseData(
        fmt=Struct,
        bindata=bytes.fromhex("000177"),
        nbytes=None,
        values=(
            Struct(datatype=0, data=1, bookend=0x77),
            [0, 1, 0x77],
        ),
        dump=Baseline(
            """
            +--------+----------+-------+-------+--------------------+
            | Offset | Access   | Value | Bytes | Format             |
            +--------+----------+-------+-------+--------------------+
            |        |          |       |       | Struct (Structure) |
            | 0      | datatype | 0     | 00    | uint8              |
            | 1      | data     | 1     | 01    | uint8              |
            | 2      | bookend  | 119   | 77    | uint8              |
            +--------+----------+-------+-------+--------------------+
            """
        ),
        excess=Baseline(
            """
            +--------+----------+----------------+-------+--------------------+
            | Offset | Access   | Value          | Bytes | Format             |
            +--------+----------+----------------+-------+--------------------+
            |        |          |                |       | Struct (Structure) |
            | 0      | datatype | 0              | 00    | uint8              |
            | 1      | data     | 1              | 01    | uint8              |
            | 2      | bookend  | 119            | 77    | uint8              |
            +--------+----------+----------------+-------+--------------------+
            | 3      |          | <excess bytes> | 99    |                    |
            +--------+----------+----------------+-------+--------------------+

            ExcessMemoryError occurred during unpack operation:

            1 unconsumed bytes
            """
        ),
        shortage=Baseline(
            """
            +--------+----------+----------------------+-------+--------------------+
            | Offset | Access   | Value                | Bytes | Format             |
            +--------+----------+----------------------+-------+--------------------+
            |        |          |                      |       | Struct (Structure) |
            | 0      | datatype | 0                    | 00    | uint8              |
            | 1      | data     | 1                    | 01    | uint8              |
            |        | bookend  | <insufficient bytes> |       | uint8              |
            +--------+----------+----------------------+-------+--------------------+

            InsufficientMemoryError occurred during unpack operation:

            1 too few bytes to unpack uint8, 1 needed, only 0 available
            """
        ),
        implementation=Baseline(
            """
            def __init__(self, *, datatype: int, data: int, bookend: int) -> None:
                self[:] = (datatype, data, bookend)

            @datatype.getter
            def datatype(self) -> int:
                return self[0]

            @datatype.setter
            def datatype(self, value: int) -> None:
                self[0] = value

            @data.getter
            def data(self) -> int:
                return self[1]

            @data.setter
            def data(self, value: int) -> None:
                self[1] = value

            @bookend.getter
            def bookend(self) -> int:
                return self[2]

            @bookend.setter
            def bookend(self, value: int) -> None:
                self[2] = value

            @classmethod
            def __pack__(cls, value, pieces: List[bytes], dump: Optional[Record] = None) -> None:
                if isinstance(value, dict):
                    value = cls._make_structure_from_dict(value)

                (m_datatype, m_data, m_bookend) = value

                if dump is None:
                    cls.datatype.fmt.__pack__(m_datatype, pieces, dump)

                    data_fmt = cls.data.fmt(value)
                    data_fmt.__pack__(m_data, pieces, dump)

                    cls.bookend.fmt.__pack__(m_bookend, pieces, dump)

                else:
                    datatype_dump = dump.add_record(access="datatype", fmt=cls.datatype.fmt)
                    cls.datatype.fmt.__pack__(m_datatype, pieces, datatype_dump)

                    data_fmt = cls.data.fmt(value)
                    data_dump = dump.add_record(access="data", fmt=data_fmt)
                    data_fmt.__pack__(m_data, pieces, data_dump)

                    bookend_dump = dump.add_record(access="bookend", fmt=cls.bookend.fmt)
                    cls.bookend.fmt.__pack__(m_bookend, pieces, bookend_dump)

            @classmethod
            def __unpack__(cls, buffer: bytes, offset: int, dump: Optional[Record] = None) -> Tuple["Struct", int]:
                structure = list.__new__(cls)
                add_member = structure.append

                if dump is None:
                    m_datatype, offset = cls.datatype.fmt.__unpack__(buffer, offset, dump)
                    add_member(m_datatype)

                    data_fmt = cls.data.fmt(structure)
                    m_data, offset = data_fmt.__unpack__(buffer, offset, dump)
                    add_member(m_data)

                    m_bookend, offset = cls.bookend.fmt.__unpack__(buffer, offset, dump)
                    add_member(m_bookend)

                else:
                    datatype_dump = dump.add_record(access="datatype", fmt=cls.datatype.fmt)
                    m_datatype, offset = cls.datatype.fmt.__unpack__(buffer, offset, datatype_dump)
                    add_member(m_datatype)

                    data_fmt = cls.data.fmt(structure)
                    data_dump = dump.add_record(access="data", fmt=data_fmt)
                    m_data, offset = data_fmt.__unpack__(buffer, offset, data_dump)
                    add_member(m_data)

                    bookend_dump = dump.add_record(access="bookend", fmt=cls.bookend.fmt)
                    m_bookend, offset = cls.bookend.fmt.__unpack__(buffer, offset, bookend_dump)
                    add_member(m_bookend)

                return structure, offset

            def __eq__(self, other: Any) -> bool:
                if isinstance(other, dict):
                    other = self._make_structure_from_dict(other)
                return list.__eq__(self, other)

            def __ne__(self, other: Any) -> bool:
                if isinstance(other, dict):
                    other = self._make_structure_from_dict(other)
                return list.__ne__(self, other)

            def __repr__(self) -> str:
                try:
                    return f"{type(self).__name__}(datatype={self.datatype!r}, data={self.data!r}, bookend={self.bookend!r})"
                except Exception:
                    return f"{type(self).__name__}()"

            """
        ),
    )


class TestProperty(Case):

    """Factory passed half baked structure instance (factory is a property method)."""

    class Struct(Structure):
        @property
        def datatype_fmt(self):
            mapping = {0: uint8, 1: uint16}
            return mapping[self[0]]

        datatype: int = member(fmt=uint8)
        data: int = member(fmt=datatype_fmt)
        bookend: int = member(fmt=uint8)

    data = CaseData(
        fmt=Struct,
        bindata=bytes.fromhex("000177"),
        nbytes=None,
        values=(
            Struct(datatype=0, data=1, bookend=0x77),
            [0, 1, 0x77],
        ),
        dump=Baseline(
            """
            +--------+----------+-------+-------+--------------------+
            | Offset | Access   | Value | Bytes | Format             |
            +--------+----------+-------+-------+--------------------+
            |        |          |       |       | Struct (Structure) |
            | 0      | datatype | 0     | 00    | uint8              |
            | 1      | data     | 1     | 01    | uint8              |
            | 2      | bookend  | 119   | 77    | uint8              |
            +--------+----------+-------+-------+--------------------+
            """
        ),
        excess=Baseline(
            """
            +--------+----------+----------------+-------+--------------------+
            | Offset | Access   | Value          | Bytes | Format             |
            +--------+----------+----------------+-------+--------------------+
            |        |          |                |       | Struct (Structure) |
            | 0      | datatype | 0              | 00    | uint8              |
            | 1      | data     | 1              | 01    | uint8              |
            | 2      | bookend  | 119            | 77    | uint8              |
            +--------+----------+----------------+-------+--------------------+
            | 3      |          | <excess bytes> | 99    |                    |
            +--------+----------+----------------+-------+--------------------+

            ExcessMemoryError occurred during unpack operation:

            1 unconsumed bytes
            """
        ),
        shortage=Baseline(
            """
            +--------+----------+----------------------+-------+--------------------+
            | Offset | Access   | Value                | Bytes | Format             |
            +--------+----------+----------------------+-------+--------------------+
            |        |          |                      |       | Struct (Structure) |
            | 0      | datatype | 0                    | 00    | uint8              |
            | 1      | data     | 1                    | 01    | uint8              |
            |        | bookend  | <insufficient bytes> |       | uint8              |
            +--------+----------+----------------------+-------+--------------------+

            InsufficientMemoryError occurred during unpack operation:

            1 too few bytes to unpack uint8, 1 needed, only 0 available
            """
        ),
        implementation=Baseline(
            """
            def __init__(self, *, datatype: int, data: int, bookend: int) -> None:
                self[:] = (datatype, data, bookend)

            @datatype.getter
            def datatype(self) -> int:
                return self[0]

            @datatype.setter
            def datatype(self, value: int) -> None:
                self[0] = value

            @data.getter
            def data(self) -> int:
                return self[1]

            @data.setter
            def data(self, value: int) -> None:
                self[1] = value

            @bookend.getter
            def bookend(self) -> int:
                return self[2]

            @bookend.setter
            def bookend(self, value: int) -> None:
                self[2] = value

            @classmethod
            def __pack__(cls, value, pieces: List[bytes], dump: Optional[Record] = None) -> None:
                if isinstance(value, dict):
                    value = cls._make_structure_from_dict(value)

                (m_datatype, m_data, m_bookend) = value

                if dump is None:
                    cls.datatype.fmt.__pack__(m_datatype, pieces, dump)

                    data_fmt = cls.data.fmt(value)
                    data_fmt.__pack__(m_data, pieces, dump)

                    cls.bookend.fmt.__pack__(m_bookend, pieces, dump)

                else:
                    datatype_dump = dump.add_record(access="datatype", fmt=cls.datatype.fmt)
                    cls.datatype.fmt.__pack__(m_datatype, pieces, datatype_dump)

                    data_fmt = cls.data.fmt(value)
                    data_dump = dump.add_record(access="data", fmt=data_fmt)
                    data_fmt.__pack__(m_data, pieces, data_dump)

                    bookend_dump = dump.add_record(access="bookend", fmt=cls.bookend.fmt)
                    cls.bookend.fmt.__pack__(m_bookend, pieces, bookend_dump)

            @classmethod
            def __unpack__(cls, buffer: bytes, offset: int, dump: Optional[Record] = None) -> Tuple["Struct", int]:
                structure = list.__new__(cls)
                add_member = structure.append

                if dump is None:
                    m_datatype, offset = cls.datatype.fmt.__unpack__(buffer, offset, dump)
                    add_member(m_datatype)

                    data_fmt = cls.data.fmt(structure)
                    m_data, offset = data_fmt.__unpack__(buffer, offset, dump)
                    add_member(m_data)

                    m_bookend, offset = cls.bookend.fmt.__unpack__(buffer, offset, dump)
                    add_member(m_bookend)

                else:
                    datatype_dump = dump.add_record(access="datatype", fmt=cls.datatype.fmt)
                    m_datatype, offset = cls.datatype.fmt.__unpack__(buffer, offset, datatype_dump)
                    add_member(m_datatype)

                    data_fmt = cls.data.fmt(structure)
                    data_dump = dump.add_record(access="data", fmt=data_fmt)
                    m_data, offset = data_fmt.__unpack__(buffer, offset, data_dump)
                    add_member(m_data)

                    bookend_dump = dump.add_record(access="bookend", fmt=cls.bookend.fmt)
                    m_bookend, offset = cls.bookend.fmt.__unpack__(buffer, offset, bookend_dump)
                    add_member(m_bookend)

                return structure, offset

            def __eq__(self, other: Any) -> bool:
                if isinstance(other, dict):
                    other = self._make_structure_from_dict(other)
                return list.__eq__(self, other)

            def __ne__(self, other: Any) -> bool:
                if isinstance(other, dict):
                    other = self._make_structure_from_dict(other)
                return list.__ne__(self, other)

            def __repr__(self) -> str:
                try:
                    return f"{type(self).__name__}(datatype={self.datatype!r}, data={self.data!r}, bookend={self.bookend!r})"
                except Exception:
                    return f"{type(self).__name__}()"

            """
        ),
    )


class TestExceptions:

    """Test exception use cases."""

    def test_bad_key_reference(self):
        """Verify exception/message if Member gets an invalid fmt_arg."""
        with pytest.raises(TypeError) as trap:

            class BadRef(Structure):  # pylint: disable=unused-variable

                """Simple variably typed member with bad reference."""

                datatype: int = member(fmt=uint8)
                data: int = member(fmt_arg="junk", fmt={0: uint8}.__getitem__)

        expected = Baseline(
            """
            'fmt_arg' must be a structure 'member()'
            """
        )

        assert wrap_message(trap.value) == expected



# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Copyright 2021 Daniel Mark Gass, see __about__.py for license information.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
"""Base class for bytes to values (and vice versa) transforms."""

from typing import Any, List, Optional, Tuple, Union

from . import exceptions
from .dump import Dump, Record


class TransformMeta(type):

    """Meta class for plum types."""

    # FUTURE - remove, this metaclass not needed functionally, just for mypy

    __nbytes__: Union[None, int]


class Transform(metaclass=TransformMeta):

    """Packable/Unpacked bytes base class."""

    __nbytes__: Union[int, None] = None

    __format_name__: str
    __hint__: str

    def __init__(self, name: str, hint: str = "Any") -> None:
        self.__format_name__ = name
        self.__hint__ = hint

    @property
    def name(self) -> str:
        """Transform format name (for repr and dump "Format" column)."""
        return self.__format_name__

    @property
    def nbytes(self) -> int:
        """Transform format size in bytes."""
        nbytes = self.__nbytes__

        if nbytes is None:
            raise exceptions.SizeError(f"{self!r} sizes vary")

        return nbytes

    def __pack__(
        self, value: Any, pieces: List[bytes], dump: Optional[Record] = None
    ) -> None:
        raise TypeError(f"{self!r} does not support pack")  # pragma: no cover

    def __unpack__(
        self, buffer: bytes, offset: int, dump: Optional[Record] = None
    ) -> Tuple[Any, int]:
        raise TypeError(f"{self!r} does not support unpack")  # pragma: no cover

    def __view__(self, buffer, offset=0):
        """Create view of bytes buffer."""
        raise TypeError(f"{self!r} does not support view()")

    def pack(self, value: Any) -> bytes:
        """Pack value as formatted bytes.

        :raises: ``PackError`` if type error, value error, etc.

        """
        pieces: List[bytes] = []
        try:
            # None -> dump
            self.__pack__(value, pieces, None)
        except Exception as exc:
            # do it over to include dump in exception message
            self.pack_and_dump(value)

            raise exceptions.ImplementationError() from exc  # pragma: no cover

        return b"".join(pieces)

    def pack_and_dump(self, value: Any) -> Tuple[bytes, Dump]:
        """Pack value as formatted bytes and produce bytes summary.

        :raises: ``PackError`` if type error, value error, etc.

        """
        dump = Dump()
        pieces: List[bytes] = []
        try:
            self.__pack__(value, pieces, dump.add_record(fmt=self.__format_name__))
        except Exception as exc:
            raise exceptions.PackError(dump=dump, exception=exc) from exc

        return b"".join(pieces), dump

    def unpack(self, buffer: bytes) -> Any:
        """Unpack value from formatted bytes.

        :raises: ``UnpackError`` if insufficient bytes, excess bytes, or value error

        """
        try:
            # None -> dump
            item, offset = self.__unpack__(buffer, 0, None)
            if buffer[offset:]:
                raise exceptions.ExcessMemoryError()
            return item

        except Exception as exc:
            # do it over to include dump in exception message
            self.unpack_and_dump(buffer)
            raise exceptions.ImplementationError() from exc  # pragma: no cover

    def unpack_and_dump(self, buffer: bytes) -> Tuple[Any, Dump]:
        """Unpack value from bytes and produce packed bytes summary.

        :raises: ``UnpackError`` if insufficient bytes, excess bytes, or value error

        """
        dmp = Dump()
        try:
            item, offset = self.__unpack__(
                buffer, 0, dmp.add_record(fmt=self.__format_name__)
            )

            extra_bytes = buffer[offset:]

            if extra_bytes:
                for i in range(0, len(extra_bytes), 16):
                    record = dmp.add_record(memory=extra_bytes[i : i + 16])
                    if not i:
                        record.separate = True
                        record.value = "<excess bytes>"

                raise exceptions.ExcessMemoryError(extra_bytes)

        except Exception as exc:
            raise exceptions.UnpackError(dmp, exc) from exc

        return item, dmp

    def view(self, buffer: bytearray, offset: int = 0) -> Any:
        """Create view of bytes buffer."""
        return self.__view__(buffer, offset)

    def __repr__(self) -> str:
        return f"<transform {self.__format_name__!r}>"



# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Copyright 2021 Daniel Mark Gass, see __about__.py for license information.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
"""Test bytes transform properties."""

from plum.bytes import BytesX


class TestDefault:

    """Test with as many left to default as possible."""

    bytesx = BytesX(name="bytesx", nbytes=4)

    def test_name(self):
        assert self.bytesx.name == "bytesx"

    def test_nbytes(self):
        assert self.bytesx.nbytes == 4

    def test_pad(self):
        assert self.bytesx.pad == b""


class TestPositional:

    """Test explicitly defined with positional argument."""

    bytesx = BytesX(4, b"\x99", "name")

    def test_hint(self):
        assert self.bytesx.__hint__ == "bytes"

    def test_name(self):
        assert self.bytesx.name == "name"

    def test_nbytes(self):
        assert self.bytesx.nbytes == 4

    def test_pad(self):
        assert self.bytesx.pad == b"\x99"


class TestKeyword(TestPositional):

    """Test explicitly defined with keyword argument."""

    bytesx = BytesX(nbytes=4, pad=b"\x99", name="name")


class TestDefaultNameHint:
    def test_greedy(self):
        xform = BytesX()
        assert xform.__hint__ == "bytes"
        assert xform.name == "bytes (greedy)"

    def test_fixed(self):
        xform = BytesX(nbytes=4)
        assert xform.__hint__ == "bytes"
        assert xform.name == "bytes (fixed)"

    def test_padded(self):
        xform = BytesX(nbytes=4, pad=b" ")
        assert xform.__hint__ == "bytes"
        assert xform.name == "bytes (fixed,padded)"



# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Copyright 2021 Daniel Mark Gass, see __about__.py for license information.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
"""Test basic features of BitFields data store transform."""

# pylint: disable=invalid-name,unexpected-keyword-arg

import sys
from enum import IntEnum

import pytest
from baseline import Baseline

from plum.bitfields import BitFields, bitfield
from plum.conformance import wrap_message
from plum.enum import EnumX


class MyEnum(IntEnum):

    """Sample IntEnum."""

    A = 1
    B = 2


class TestInit:

    """Test constructor."""

    class MyBits(BitFields, nbytes=2, default=0x0):
        """Sample BitFields subclass."""

        f1: int = bitfield(lsb=0, size=8, default=0xAB)
        f2: MyEnum = bitfield(lsb=8, size=4)
        f3: bool = bitfield(lsb=12, size=1)
        f4: bool = bitfield(lsb=13, size=1)

    def test_invalid_arg_type(self):
        """Test invalid argument type passed to constructor."""
        expected_message = Baseline(
            """
            int() argument must be a string, a bytes-like object or a real number, not 'list'
            """
        )

        with pytest.raises(TypeError) as trap:
            self.MyBits(f1=[1], f2=0, f3=True, f4=True)

        actual_message = str(trap.value)

        if sys.version_info < (3, 10):
            actual_message = actual_message.replace("number", "real number")

        assert actual_message == expected_message


class TestInvalidInitValues:

    """Test out of range values passed to constructor."""

    class Struct(BitFields):
        f1 = bitfield(size=4, default=0)
        f2 = bitfield(size=4, signed=True, default=0)

    unsigned_message = Baseline(
        """
        bit field 'f1' requires 0 <= number <= 15
        """
    )

    signed_message = Baseline(
        """
        bit field 'f2' requires -8 <= number <= 7
        """
    )

    def test_too_small_for_signed(self):
        with pytest.raises(ValueError) as trap:
            self.Struct(f2=-9)

        assert wrap_message(trap.value) == self.signed_message

    def test_too_small_for_unsigned(self):
        with pytest.raises(ValueError) as trap:
            self.Struct(f1=-1)

        assert wrap_message(trap.value) == self.unsigned_message

    def test_too_big_for_signed(self):
        with pytest.raises(ValueError) as trap:
            self.Struct(f2=8)

        assert wrap_message(trap.value) == self.signed_message

    def test_too_big_for_unsigned(self):
        with pytest.raises(ValueError) as trap:
            self.Struct(f1=16)

        assert wrap_message(trap.value) == self.unsigned_message


class TestInvalidAttributeValues:

    """Test out of range attribute values."""

    class Struct(BitFields):
        f1 = bitfield(size=4, default=0)
        f2 = bitfield(size=4, signed=True, default=0)

    unsigned_message = Baseline(
        """
        bit field 'f1' requires 0 <= number <= 15
        """
    )

    signed_message = Baseline(
        """
        bit field 'f2' requires -8 <= number <= 7
        """
    )

    def test_too_small_for_signed(self):
        with pytest.raises(ValueError) as trap:
            self.Struct().f2 = -9

        assert wrap_message(trap.value) == self.signed_message

    def test_too_small_for_unsigned(self):
        with pytest.raises(ValueError) as trap:
            self.Struct().f1 = -1

        assert wrap_message(trap.value) == self.unsigned_message

    def test_too_big_for_signed(self):
        with pytest.raises(ValueError) as trap:
            self.Struct().f2 = 8

        assert wrap_message(trap.value) == self.signed_message

    def test_too_big_for_unsigned(self):
        with pytest.raises(ValueError) as trap:
            self.Struct().f1 = 16

        assert wrap_message(trap.value) == self.unsigned_message


class ForCompare(BitFields, nbytes=1, default=0x0, ignore=0xC0):

    """Sample BitFields subclass with ignored fields."""

    f1: int = bitfield(lsb=0, size=4)
    f2: int = bitfield(lsb=4, size=2, ignore=True)
    f3: int = bitfield(lsb=6, size=2)  # gets ignored from class 'ignore'


class TestEquality:

    """Test __eq__ and __ne__ operators.

    Test common normalization algorithms that implement ignore mechanisms.
    Rely on TestCompares to thoroughly check support for each of the other
    comparison methods (e.g. __le__, __lt__, etc.).

    """

    # pylint: disable=unneeded-not

    def test_int(self):
        """Verify when other is int, nothing ignored"""
        assert ForCompare(f1=1, f2=2, f3=2) == 0xA1

    def test_ignore_field(self):
        """Verify when other is same type as self, field marked as ignore doesn't matter."""
        assert ForCompare(f1=1, f2=2, f3=2) == ForCompare(f1=1, f2=0, f3=2)
        assert not ForCompare(f1=0, f2=2, f3=2) == ForCompare(f1=1, f2=2, f3=2)

    def test_ignore_mask(self):
        """Verify when other is same type as self, subclass ignore mask applies."""
        assert ForCompare(f1=1, f2=2, f3=2) == ForCompare(f1=1, f2=2, f3=3)
        assert not ForCompare(f1=0, f2=2, f3=2) == ForCompare(f1=1, f2=2, f3=2)

    def test_dict(self):
        """Verify support for other being a dict."""
        assert ForCompare(f1=1, f2=2, f3=2) == dict(f1=1, f2=2, f3=2)


class TestCompares:

    """Verify comparison operators.

    Assume all operators implemented with a common normalization
    algorithm. Only spot check each operator and rely on TestEquality
    test cases to validate the common normalization algorithm.

    """

    # pylint: disable=unneeded-not

    def test_lt(self):
        """Spot check 'lt' operator."""
        assert ForCompare(f1=1, f2=2, f3=2) < 0xA2
        assert not ForCompare(f1=1, f2=2, f3=2) < 0xA0

    def test_le(self):
        """Spot check 'lt' operator."""
        assert ForCompare(f1=1, f2=2, f3=2) <= 0xA1
        assert not ForCompare(f1=1, f2=2, f3=2) <= 0xA0

    def test_eq(self):
        """Spot check 'eq' operator."""
        assert ForCompare(f1=1, f2=2, f3=2) == 0xA1
        assert not ForCompare(f1=1, f2=2, f3=2) == 0xA0

    def test_ne(self):
        """Spot check 'ne' operator."""
        assert ForCompare(f1=1, f2=2, f3=2) != 0xA0
        assert not ForCompare(f1=1, f2=2, f3=2) != 0xA1

    def test_gt(self):
        """Spot check 'gt' operator."""
        assert ForCompare(f1=1, f2=2, f3=2) > 0xA0
        assert not ForCompare(f1=1, f2=2, f3=2) > 0xA2

    def test_ge(self):
        """Spot check 'ge' operator."""
        assert ForCompare(f1=1, f2=2, f3=2) >= 0xA1
        assert not ForCompare(f1=1, f2=2, f3=2) >= 0xA2


class TestCompareValueError:

    """Test when other can not be converted to an int."""

    # pylint: disable=unneeded-not

    def test_eq(self):
        assert not ForCompare(f1=2, f2=2, f3=2) == "not valid"

    def test_ne(self):
        assert ForCompare(f1=2, f2=2, f3=2) != "not valid"


class SimpleBits(BitFields, nbytes=1):

    """Sample BitFields subclass."""

    f1: int = bitfield(lsb=0, size=8, default=0)


class TestArithmetic:

    """Verify arithmetic operators.

    Assume all operators implemented with a common normalization
    algorithm. Only spot check each operator.

    """

    # pylint: disable=unidiomatic-typecheck,too-many-public-methods

    def test_add(self):
        """Spot check '+' operator."""
        x = SimpleBits.from_int(0) + 1
        assert x == 1
        assert type(x) is int

    def test_sub(self):
        """Spot check '-' operator."""
        x = SimpleBits.from_int(3) - 1
        assert x == 2
        assert type(x) is int

    def test_mul(self):
        """Spot check '*' operator."""
        x = SimpleBits.from_int(3) * 4
        assert x == 12
        assert type(x) is int

    def test_truediv(self):
        """Spot check '/' operator."""
        x = SimpleBits.from_int(9) / 2
        assert x == 4.5
        assert type(x) is float

    def test_floordiv(self):
        """Spot check '//' operator."""
        x = SimpleBits.from_int(9) // 2
        assert x == 4
        assert type(x) is int

    def test_mod(self):
        """Spot check '%' operator."""
        x = SimpleBits.from_int(9) % 2
        assert x == 1
        assert type(x) is int

    def test_divmod(self):
        """Spot check divmod() operator."""
        div, mod = divmod(SimpleBits.from_int(9), 2)
        assert div == 4
        assert mod == 1
        assert type(div) is int
        assert type(mod) is int

    def test_pow(self):
        """Spot check '**' operator."""
        x = SimpleBits.from_int(2) ** 3
        assert x == 8
        assert type(x) is int

        x = pow(SimpleBits.from_int(2), 3, 5)
        assert x == 3
        assert type(x) is int

    def test_radd(self):
        """Spot check '+' operator (right hand size)."""
        x = 1 + SimpleBits.from_int(0)
        assert x == 1
        assert type(x) is int

    def test_rsub(self):
        """Spot check '-' operator (right hand size)."""
        x = 3 - SimpleBits.from_int(1)
        assert x == 2
        assert type(x) is int

    def test_rmul(self):
        """Spot check '*' operator (right hand size)."""
        x = 4 * SimpleBits.from_int(3)
        assert x == 12
        assert type(x) is int

    def test_rtruediv(self):
        """Spot check '/' operator (right hand size)."""
        x = 9 / SimpleBits.from_int(2)
        assert x == 4.5
        assert type(x) is float

    def test_rfloordiv(self):
        """Spot check '//' operator (right hand size)."""
        x = 9 // SimpleBits.from_int(2)
        assert x == 4
        assert type(x) is int

    def test_rmod(self):
        """Spot check '%' operator (right hand size)."""
        x = 9 % SimpleBits.from_int(2)
        assert x == 1
        assert type(x) is int

    def test_rdivmod(self):
        """Spot check divmod() operator (right hand size)."""
        div, mod = divmod(9, SimpleBits.from_int(2))
        assert div == 4
        assert mod == 1
        assert type(div) is int
        assert type(mod) is int

    def test_rpow(self):
        """Spot check '**' operator (right hand size)."""
        x = 2 ** SimpleBits.from_int(3)
        assert x == 8
        assert type(x) is int

        # Note, Python does not support this:
        # x = pow(2, SimpleBits.from_int(3), 5)
        # assert x == 3
        # assert type(x) is int

    def test_iadd(self):
        """Spot check '+' operator (in place)."""
        x = SimpleBits.from_int(0)
        x += 1
        assert x == 1
        assert type(x) is int

    def test_isub(self):
        """Spot check '-' operator (in place)."""
        x = SimpleBits.from_int(3)
        x -= 1
        assert x == 2
        assert type(x) is int

    def test_imul(self):
        """Spot check '*' operator (in place)."""
        x = SimpleBits.from_int(3)
        x *= 4
        assert x == 12
        assert type(x) is int

    def test_itruediv(self):
        """Spot check '/' operator (in place)."""
        x = SimpleBits.from_int(9)
        x /= 2
        assert x == 4.5
        assert type(x) is float

    def test_ifloordiv(self):
        """Spot check '//' operator (in place)."""
        x = SimpleBits.from_int(9)
        x //= 2
        assert x == 4
        assert type(x) is int

    def test_imod(self):
        """Spot check '%' operator (in place)."""
        x = SimpleBits.from_int(9)
        x %= 2
        assert x == 1
        assert type(x) is int

    def test_neg(self):
        """Spot check neg() operator."""
        x = -SimpleBits.from_int(9)  # pylint: disable=invalid-unary-operand-type
        assert x == -9
        assert type(x) is int

    def test_pos(self):
        """Spot check neg() operator."""
        x = +SimpleBits.from_int(9)  # pylint: disable=invalid-unary-operand-type
        assert x == 9
        assert type(x) is int

    def test_abs(self):
        """Spot check abs() operator."""
        x = abs(SimpleBits.from_int(9))
        assert x == 9
        assert type(x) is int

    def test_invert(self):
        """Spot check '~' operator."""
        x = ~SimpleBits.from_int(1)  # pylint: disable=invalid-unary-operand-type
        assert x == -2
        assert type(x) is int

    def test_float(self):
        """Spot check float() operator."""
        x = float(SimpleBits.from_int(9))
        assert x == 9.0

    def test_index(self):
        """Spot check __index__ method."""
        numbers = list(range(10))
        # pylint: disable=invalid-sequence-index
        assert numbers[SimpleBits.from_int(1)] == numbers[1]

    def test_round(self):
        """Spot check round() support."""
        x = round(SimpleBits.from_int(1))
        assert x == 1
        assert type(x) is int


class TestBitLogic:
    """Verify bit shift and logical operators.

    Assume all operators implemented with a common normalization
    algorithm. Only spot check each operator.

    """

    # pylint: disable=unidiomatic-typecheck

    def test_lshift(self):
        """Spot check '<<' operator."""
        x = SimpleBits.from_int(1) << 2
        assert x == 1 << 2
        assert type(x) is int

    def test_rshift(self):
        """Spot check '>>' operator."""
        x = SimpleBits.from_int(1 << 2) >> 2
        assert x == 1
        assert type(x) is int

    def test_and(self):
        """Spot check '&' operator."""
        x = SimpleBits.from_int(3) & 2
        assert x == 2
        assert type(x) is int

    def test_xor(self):
        """Spot check '^' operator."""
        x = SimpleBits.from_int(0b1010) ^ 0b1100
        assert x == 0b0110
        assert type(x) is int

    def test_or(self):
        """Spot check '|' operator."""
        x = SimpleBits.from_int(0b010) | 0b001
        assert x == 0b011
        assert type(x) is int

    def test_rlshift(self):
        """Spot check '<<' operator (right hand size)."""
        x = 1 << SimpleBits.from_int(2)
        assert x == 1 << 2
        assert type(x) is int

    def test_rrshift(self):
        """Spot check '>>' operator (right hand size)."""
        x = (1 << 2) >> SimpleBits.from_int(2)
        assert x == 1
        assert type(x) is int

    def test_rand(self):
        """Spot check '&' operator (right hand size)."""
        x = 2 & SimpleBits.from_int(3)
        assert x == 2
        assert type(x) is int

    def test_rxor(self):
        """Spot check '^' operator (right hand size)."""
        x = 0b1100 ^ SimpleBits.from_int(0b1010)
        assert x == 0b0110
        assert type(x) is int

    def test_ror(self):
        """Spot check '|' operator (right hand size)."""
        x = 0b001 | SimpleBits.from_int(0b010)
        assert x == 0b011
        assert type(x) is int

    def test_ilshift(self):
        """Spot check '<<' operator (in place)."""
        x = SimpleBits.from_int(1)
        x <<= 2
        assert x == 1 << 2
        assert type(x) is int

    def test_irshift(self):
        """Spot check '>>' operator (in place)."""
        x = SimpleBits.from_int(1 << 2)
        x >>= 2
        assert x == 1
        assert type(x) is int

    def test_iand(self):
        """Spot check '&' operator (in place)."""
        x = SimpleBits.from_int(3)
        x &= 2
        assert x == 2
        assert type(x) is int

    def test_ixor(self):
        """Spot check '^' operator (in place)."""
        x = SimpleBits.from_int(0b1010)
        x ^= 0b1100
        assert x == 0b0110
        assert type(x) is int

    def test_ior(self):
        """Spot check '|' operator (in place)."""
        x = SimpleBits.from_int(0b010)
        x |= 0b001
        assert x == 0b011
        assert type(x) is int


class TestHash:

    """Test support of hash()."""

    def test_hash(self):
        """Verify hash does not include ignored fields."""
        assert hash(ForCompare(f1=1, f2=1, f3=1)) == hash(1)


class TestBool:

    """Test conversion with bool()."""

    def test_bool(self):
        """Verify bool does not consider ignored fields."""
        assert ForCompare(f1=1, f2=0, f3=0)
        assert not ForCompare(f1=0, f2=1, f3=1)


class ForSigned(BitFields, nbytes=1):

    """Sample BitFields subclass with signed field."""

    f1: int = bitfield(lsb=0, size=2)
    f2: int = bitfield(lsb=2, size=4, signed=True)
    f3: int = bitfield(lsb=6, size=2)


class TestSigned:

    """Test signed BitField member."""

    def test_get(self):
        """Verify get access."""
        assert ForSigned.from_int(0xC3).f2 == 0
        assert ForSigned.from_int(0x3C).f2 == -1
        assert ForSigned.from_int(0xE3).f2 == -8

    def test_valid_set(self):
        """Verify set access."""
        bitfields = ForSigned.from_int(0)
        bitfields.f2 = -1
        assert int(bitfields) == 0x3C

        bitfields = ForSigned.from_int(0)
        bitfields.f2 = 7
        assert int(bitfields) == 0x1C

        bitfields = ForSigned.from_int(0xC3)
        bitfields.f2 = -8
        assert int(bitfields) == 0xE3

        bitfields = ForSigned.from_int(0xFF)
        bitfields.f2 = 0
        assert int(bitfields) == 0xC3

    def test_invalid_set(self):
        """Test value out of range."""
        expected = Baseline(
            """
            bit field 'f2' requires -8 <= number <= 7
            """
        )

        bitfields = ForSigned.from_int(0)

        with pytest.raises(ValueError) as trap:
            bitfields.f2 = 8
        assert wrap_message(trap.value) == expected

        with pytest.raises(ValueError) as trap:
            bitfields.f2 = -9
        assert wrap_message(trap.value) == expected


class ForUnsigned(BitFields, nbytes=1):

    """Sample BitFields subclass with signed field."""

    f1: int = bitfield(lsb=0, size=2)
    f2: int = bitfield(lsb=2, size=4)
    f3: int = bitfield(lsb=6, size=2)


class TestUnsigned:

    """Test unsigned bit field member."""

    def test_get(self):
        """Verify get access."""
        assert ForUnsigned.from_int(0xC3).f2 == 0
        assert ForUnsigned.from_int(0xC7).f2 == 1
        assert ForUnsigned.from_int(0xE3).f2 == 8
        assert ForUnsigned.from_int(0x3C).f2 == 0xF

    def test_valid_set(self):
        """Verify set access."""
        bitfields = ForUnsigned.from_int(0)
        bitfields.f2 = 0xF
        assert int(bitfields) == 0x3C

        bitfields = ForUnsigned.from_int(0xFF)
        bitfields.f2 = 0
        assert int(bitfields) == 0xC3

        bitfields = ForUnsigned.from_int(0x0)
        bitfields.f2 = 1
        assert int(bitfields) == 4

        bitfields = ForUnsigned.from_int(0x0)
        bitfields.f2 = 8
        assert int(bitfields) == 0x20

    def test_invalid_set(self):
        """Test value out of range."""
        expected = Baseline(
            """
            bit field 'f2' requires 0 <= number <= 15
            """
        )

        bitfields = ForUnsigned.from_int(0)

        with pytest.raises(ValueError) as trap:
            bitfields.f2 = -1
        assert wrap_message(trap.value) == expected

        with pytest.raises(ValueError) as trap:
            bitfields.f2 = 16
        assert wrap_message(trap.value) == expected


class ForStrictEnum(BitFields, nbytes=1):

    """Sample BitFields subclass with enumeration field."""

    f1: int = bitfield(lsb=0, size=2)
    f2: MyEnum = bitfield(lsb=2, size=4, typ=MyEnum)
    f3: int = bitfield(lsb=6, size=2)


class TestStrictEnum:

    """Test enumeration bit field member."""

    def test_happy_get(self):
        """Verify get access succeeds with valid value."""
        assert ForStrictEnum.from_int(1 << 2).f2 is MyEnum.A

    def test_unhappy_get(self):
        """Verify get access fails when invalid value."""

        expected_message = Baseline(
            """
            0 is not a valid MyEnum
            """
        )

        with pytest.raises(ValueError) as trap:
            # pylint: disable=expression-not-assigned
            ForStrictEnum.from_int(0).f2

        assert str(trap.value) == expected_message

    def test_happy_set(self):
        """Verify set access succeeds with valid value."""
        bitfields = ForStrictEnum.from_int(0)
        bitfields.f2 = MyEnum.A
        assert int(bitfields) == 1 << 2

    def test_unhappy_set(self):
        """Verify set access fails with invalid value."""
        expected_message = Baseline(
            """
            0 is not a valid MyEnum
            """
        )

        bitfields = ForStrictEnum.from_int(0)

        with pytest.raises(ValueError) as trap:
            bitfields.f2 = 0

        assert str(trap.value) == expected_message


class ForTolerantEnum(BitFields, nbytes=1):

    """Sample BitFields subclass with enumeration field."""

    f1: int = bitfield(lsb=0, size=2)
    f2: int = bitfield(
        lsb=2, size=4, typ=EnumX(name="myenum", enum=MyEnum, strict=False)
    )
    f3: int = bitfield(lsb=6, size=2)


class TestTolerantEnum:
    """Test enumeration bit field member."""

    def test_happy_get(self):
        """Verify get access succeeds with valid value."""
        assert ForTolerantEnum.from_int(1 << 2).f2 is MyEnum.A

    def test_tolerant_get(self):
        """Verify get access fails when invalid value."""
        assert ForTolerantEnum.from_int(0).f2 == 0

    def test_happy_set(self):
        """Verify set access succeeds with valid value."""
        bitfields = ForTolerantEnum.from_int(0)
        bitfields.f2 = MyEnum.A
        assert int(bitfields) == 1 << 2

    def test_tolerant_set(self):
        """Verify set access fails with invalid value."""
        bitfields = ForTolerantEnum.from_int(0)
        bitfields.f2 = 3
        assert int(bitfields) == 3 << 2

    def test_happy_dump(self):
        expected_dump = Baseline(
            """
            +--------+--------+----------+----------+-----------------------------+
            | Offset | Access | Value    | Bytes    | Format                      |
            +--------+--------+----------+----------+-----------------------------+
            | 0      |        | 4        | 04       | ForTolerantEnum (BitFields) |
            |  [0:2] | f1     | 0        | ......00 | int                         |
            |  [2:6] | f2     | MyEnum.A | ..0001.. | myenum                      |
            |  [6:8] | f3     | 0        | 00...... | int                         |
            +--------+--------+----------+----------+-----------------------------+
            """
        )
        assert str(ForTolerantEnum.from_int(1 << 2).dump) == expected_dump

    def test_tolerant_dump(self):
        expected_dump = Baseline(
            """
            +--------+--------+-------+----------+-----------------------------+
            | Offset | Access | Value | Bytes    | Format                      |
            +--------+--------+-------+----------+-----------------------------+
            | 0      |        | 0     | 00       | ForTolerantEnum (BitFields) |
            |  [0:2] | f1     | 0     | ......00 | int                         |
            |  [2:6] | f2     | 0     | ..0000.. | myenum                      |
            |  [6:8] | f3     | 0     | 00...... | int                         |
            +--------+--------+-------+----------+-----------------------------+
            """
        )
        assert str(ForTolerantEnum.from_int(0).dump) == expected_dump


class TestAsDict:

    """Test asdict() method."""

    class MyBits(BitFields):
        """Sample BitFields subclass."""

        f1 = bitfield(size=4)
        f2 = bitfield(size=4)

    def test_method(self):
        sample = self.MyBits(f1=1, f2=2)

        assert sample.asdict() == {"f1": 1, "f2": 2}



# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Copyright 2021 Daniel Mark Gass, see __about__.py for license information.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
"""Test base view basic methods and behaviors."""

from baseline import Baseline
from plum.int import IntX
from plum.structure import Structure, member

uint8 = IntX(name="uint8", nbytes=1)
uint16 = IntX(name="uint16", nbytes=2)


class SampleStruct(Structure):

    """Sample structure."""

    m1: int = member(fmt=uint8)


class TestCast:

    """Test base view cast() method."""

    def test_cast(self):
        """Test base view cast() method."""
        buffer = bytearray([0, 1, 2, 3])
        view_16 = uint16.view(buffer, offset=1)
        view_8 = view_16.cast(uint8)
        assert view_8 == 0x1
        assert view_16 == 0x0201
        view_8.set(0x99)
        assert view_8 == 0x99
        assert view_16 == 0x0299


class TestStr:

    """Test base view __str__() method."""

    def test_get_success(self):
        """Test str() conversion produces str() of datatype get()."""
        buffer = bytearray([0, 1, 2])
        view_8 = uint8.view(buffer, offset=1)
        assert str(view_8) == "1"

    def test_get_fail(self):
        """Test str() conversion produces representation when get() fails."""
        buffer = bytearray([0, 1, 2])
        view_8 = uint8.view(buffer, offset=3)
        assert str(view_8) == "<view at 0x3>"


class TestRepr:

    """Test base view __repr__() method."""

    def test_get_success(self):
        """Test repr() produces representation with value when get() successful."""
        buffer = bytearray([0, 1, 2])
        view_8 = uint8.view(buffer, offset=1)
        expected_repr = Baseline(
            """
            <uint8 view at 0x1: 1>
            """
        )
        assert repr(view_8) == expected_repr

    def test_get_fail_number(self):
        """Test repr() produces representation without value when get() fails.

        Test number method.

        """
        buffer = bytearray([0, 1, 2])
        view_8 = uint8.view(buffer, offset=3)
        assert repr(view_8) == "<view at 0x3>"

    def test_get_fail_base(self):
        """Test repr() produces representation without value when get() fails.

        Test base View class method.
        """
        buffer = bytearray([0, 1, 2])
        view_8 = SampleStruct.view(buffer, offset=3)
        expected_repr = Baseline(
            """
            <view at 0x3>
            """
        )
        assert repr(view_8) == expected_repr


class TestProperties:

    """Test base view properties."""

    def test_nbytes(self):
        assert SampleStruct.view(b"\x00").nbytes == 1



# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Copyright 2021 Daniel Mark Gass, see __about__.py for license information.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
"""Test special functionality of plum exception types."""

import pytest
from baseline import Baseline

from plum.conformance import wrap_message
from plum.exceptions import ImplementationError, PackError, UnpackError
from plum.littleendian import uint8
from plum.structure import Structure
from plum.utilities import pack, unpack


class TestImplementationError:

    """Test ImplementationError() exception type."""

    def test_custom_message(self):
        """Test uses message specified."""
        exc = ImplementationError("Hello World!")
        assert str(exc) == "Hello World!"

    def test_default_message(self):
        """Test uses a pre-canned message if no message specified."""
        exc = ImplementationError()

        expected = Baseline(
            """
        One of the plum types used in the pack/unpack operation contains an
        implementation error. The operation generated an exception when first
        performed without a dump (for efficiency). But when the operation was
        repeated with a dump (for a better exception message) the exception
        did not re-occur. Please report the inconsistent behavior to the type
        developer.
        """
        )

        assert wrap_message(str(exc)) == expected


class TestPackExceptions:

    """Test invalid format and value combinations when packing."""

    def test_value_not_dict(self):
        """Test value is not a dictionary when format was dict."""
        expected_message = Baseline(
            """


            +--------+-----------+-------+-------+--------+
            | Offset | Access    | Value | Bytes | Format |
            +--------+-----------+-------+-------+--------+
            |        | ['outer'] | 0     |       | dict   |
            +--------+-----------+-------+-------+--------+

            TypeError occurred during pack operation:

            invalid value, expected dict, got 0
            """
        )

        with pytest.raises(PackError) as trap:
            pack(dict(outer=0), {"outer": {"inner": uint8}})

        assert str(trap.value) == expected_message

    def test_value_not_list_like(self):
        """Test value is not a list or tuple when format called for it."""
        expected_message = Baseline(
            """


            +--------+-------+-------+---------------+
            | Offset | Value | Bytes | Format        |
            +--------+-------+-------+---------------+
            |        | 0     |       | tuple or list |
            +--------+-------+-------+---------------+

            TypeError occurred during pack operation:

            invalid value, expected list or tuple, got 0
            """
        )

        with pytest.raises(PackError) as trap:
            pack(0, [[uint8]])

        assert str(trap.value) == expected_message

    def test_fmt_not_valid_value_missing(self):
        """Test format is not a DataMeta when value reported missing."""
        with pytest.raises(PackError) as trap:
            pack({"m1": 0}, {"m1": uint8, "m2": "junk"})

        expected_message = Baseline(
            """


            +--------+--------+-----------+-------+--------+
            | Offset | Access | Value     | Bytes | Format |
            +--------+--------+-----------+-------+--------+
            | 0      | ['m1'] | 0         | 00    | uint8  |
            |        | ['m2'] | (missing) |       |        |
            +--------+--------+-----------+-------+--------+

            TypeError occurred during pack operation:

            missing value: 'm2'
            """
        )

        assert str(trap.value) == expected_message

    def test_multiple_values_missing(self):
        """Test when value has multiple missing members."""
        with pytest.raises(PackError) as trap:
            pack({"m1": 0}, {"m1": uint8, "m2": uint8, "m3": uint8})

        expected_message = Baseline(
            """


            +--------+--------+-----------+-------+--------+
            | Offset | Access | Value     | Bytes | Format |
            +--------+--------+-----------+-------+--------+
            | 0      | ['m1'] | 0         | 00    | uint8  |
            |        | ['m2'] | (missing) |       | uint8  |
            |        | ['m3'] | (missing) |       | uint8  |
            +--------+--------+-----------+-------+--------+

            TypeError occurred during pack operation:

            missing values: 'm2', 'm3'
            """
        )

        assert str(trap.value) == expected_message

    def test_too_many_arg_value(self):
        """Test too many arg values present when simple DataMeta format."""
        expected_message = Baseline(
            """


            +--------+--------+-------+-------+--------------+
            | Offset | Access | Value | Bytes | Format       |
            +--------+--------+-------+-------+--------------+
            | 0      | [0]    | 0     | 00    | uint8        |
            | 1      | [1]    | 1     | 01    | uint8        |
            +--------+--------+-------+-------+--------------+
            |        | [2]    | 2     |       | (unexpected) |
            +--------+--------+-------+-------+--------------+

            TypeError occurred during pack operation:

            3 values given, expected 2
            """
        )

        with pytest.raises(PackError) as trap:
            pack((0, 1, 2), (uint8, uint8))

        assert str(trap.value) == expected_message

    def test_too_few_list_values(self):
        """Test too few list values present."""
        expected_message = Baseline(
            """


            +--------+--------+-----------+-------+-----------+
            | Offset | Access | Value     | Bytes | Format    |
            +--------+--------+-----------+-------+-----------+
            | 0      | [0]    | 0         | 00    | uint8     |
            |        | [1]    | (missing) |       | uint8     |
            |        | [2]    | (missing) |       | Structure |
            +--------+--------+-----------+-------+-----------+

            TypeError occurred during pack operation:

            1 value given, expected 3
            """
        )

        with pytest.raises(PackError) as trap:
            pack((0,), (uint8, uint8, Structure))

        assert str(trap.value) == expected_message

    def test_bad_format_type(self):
        """Test fmt is a type that doesn't support plum pack protocol."""
        expected_message = Baseline(
            """


            +--------+-------+-------+---------------+
            | Offset | Value | Bytes | Format        |
            +--------+-------+-------+---------------+
            |        | 0     |       | int (invalid) |
            +--------+-------+-------+---------------+

            TypeError occurred during pack operation:

            bad item format, must be a packable data type/transform (or a dict, list, or tuple of them)
            """
        )

        with pytest.raises(PackError) as trap:
            pack(0, int)

        assert str(trap.value) == expected_message

    def test_no_extras(self):
        """Test exception with no dump or embedded exception."""
        assert str(PackError("Hello.")) == "Hello."


class TestUnpackExceptions:

    """Test invalid format and value combinations when unpacking."""

    def test_bad_fmt(self):
        """Test fmt is an invalid type."""
        with pytest.raises(UnpackError) as trap:
            unpack(0, b"")

        expected_message = Baseline(
            """


            +--------+-------+-------+--------+
            | Offset | Value | Bytes | Format |
            +--------+-------+-------+--------+
            |        |       |       | 0      |
            +--------+-------+-------+--------+

            TypeError occurred during unpack operation:

            bad item format, must be a packable data type/transform (or a dict, list, or tuple of them)
            """
        )

        assert str(trap.value) == expected_message

    def test_bad_nested_fmt(self):
        """Test nested fmt is an invalid type."""
        with pytest.raises(UnpackError) as trap:
            unpack([0], b"")

        expected_message = Baseline(
            """


            +--------+--------+-------+-------+--------+
            | Offset | Access | Value | Bytes | Format |
            +--------+--------+-------+-------+--------+
            |        | [0]    |       |       | 0      |
            +--------+--------+-------+-------+--------+

            TypeError occurred during unpack operation:

            bad item format, must be a packable data type/transform (or a dict, list, or tuple of them)
            """
        )

        assert str(trap.value) == expected_message



# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Copyright 2021 Daniel Mark Gass, see __about__.py for license information.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
"""Test structure data store transform API conformance."""

# pylint: disable=unexpected-keyword-arg

from baseline import Baseline

from plum.conformance import Case, CaseData
from plum.littleendian import uint8, uint16
from plum.structure import Structure, member


class Custom(Structure):

    """Sample Structure class."""

    m0: int = member(fmt=uint8)
    m1: int = member(fmt=uint16)


class TestConformance(Case):

    """Test structure data store transform conformance."""

    data = CaseData(
        fmt=Custom,
        bindata=b"\x01\x02\x00",
        nbytes=3,
        values=(Custom(m0=1, m1=2), [1, 2], (1, 2), dict(m0=1, m1=2)),
        dump=Baseline(
            """
            +--------+--------+-------+-------+--------------------+
            | Offset | Access | Value | Bytes | Format             |
            +--------+--------+-------+-------+--------------------+
            |        |        |       |       | Custom (Structure) |
            | 0      | m0     | 1     | 01    | uint8              |
            | 1      | m1     | 2     | 02 00 | uint16             |
            +--------+--------+-------+-------+--------------------+
            """
        ),
        excess=Baseline(
            """
            +--------+--------+----------------+-------+--------------------+
            | Offset | Access | Value          | Bytes | Format             |
            +--------+--------+----------------+-------+--------------------+
            |        |        |                |       | Custom (Structure) |
            | 0      | m0     | 1              | 01    | uint8              |
            | 1      | m1     | 2              | 02 00 | uint16             |
            +--------+--------+----------------+-------+--------------------+
            | 3      |        | <excess bytes> | 99    |                    |
            +--------+--------+----------------+-------+--------------------+

            ExcessMemoryError occurred during unpack operation:

            1 unconsumed bytes
            """
        ),
        shortage=Baseline(
            """
            +--------+--------+----------------------+-------+--------------------+
            | Offset | Access | Value                | Bytes | Format             |
            +--------+--------+----------------------+-------+--------------------+
            |        |        |                      |       | Custom (Structure) |
            | 0      | m0     | 1                    | 01    | uint8              |
            | 1      | m1     | <insufficient bytes> | 02    | uint16             |
            +--------+--------+----------------------+-------+--------------------+

            InsufficientMemoryError occurred during unpack operation:

            1 too few bytes to unpack uint16, 2 needed, only 1 available
            """
        ),
        implementation=Baseline(
            """
            def __init__(self, *, m0: int, m1: int) -> None:
                self[:] = (m0, m1)

            @m0.getter
            def m0(self) -> int:
                return self[0]

            @m0.setter
            def m0(self, value: int) -> None:
                self[0] = value

            @m1.getter
            def m1(self) -> int:
                return self[1]

            @m1.setter
            def m1(self, value: int) -> None:
                self[1] = value

            @classmethod
            def __pack__(cls, value, pieces: List[bytes], dump: Optional[Record] = None) -> None:
                if isinstance(value, dict):
                    value = cls._make_structure_from_dict(value)

                (m_m0, m_m1) = value

                if dump is None:
                    cls.m0.fmt.__pack__(m_m0, pieces, dump)

                    cls.m1.fmt.__pack__(m_m1, pieces, dump)

                else:
                    m0_dump = dump.add_record(access="m0", fmt=cls.m0.fmt)
                    cls.m0.fmt.__pack__(m_m0, pieces, m0_dump)

                    m1_dump = dump.add_record(access="m1", fmt=cls.m1.fmt)
                    cls.m1.fmt.__pack__(m_m1, pieces, m1_dump)

            @classmethod
            def __unpack__(cls, buffer: bytes, offset: int, dump: Optional[Record] = None) -> Tuple["Custom", int]:
                structure = list.__new__(cls)

                if dump is None:
                    m_m0, offset = cls.m0.fmt.__unpack__(buffer, offset, dump)

                    m_m1, offset = cls.m1.fmt.__unpack__(buffer, offset, dump)

                else:
                    m0_dump = dump.add_record(access="m0", fmt=cls.m0.fmt)
                    m_m0, offset = cls.m0.fmt.__unpack__(buffer, offset, m0_dump)

                    m1_dump = dump.add_record(access="m1", fmt=cls.m1.fmt)
                    m_m1, offset = cls.m1.fmt.__unpack__(buffer, offset, m1_dump)

                structure[:] = (m_m0, m_m1)

                return structure, offset

            def __eq__(self, other: Any) -> bool:
                if isinstance(other, dict):
                    other = self._make_structure_from_dict(other)
                return list.__eq__(self, other)

            def __ne__(self, other: Any) -> bool:
                if isinstance(other, dict):
                    other = self._make_structure_from_dict(other)
                return list.__ne__(self, other)

            def __repr__(self) -> str:
                try:
                    return f"{type(self).__name__}(m0={self.m0!r}, m1={self.m1!r})"
                except Exception:
                    return f"{type(self).__name__}()"

            """
        ),
    )



# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Copyright 2022 Daniel Mark Gass, see __about__.py for license information.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
"""Structure of uniquely named and typed members data store and transform."""

from typing import Any, Dict, Optional, Tuple, Union

from ..data import Data
from .bitfield_member import bitfield_member
from .dimmed_member import dimmed_member
from .member import member, Member
from .meta import StructureMeta
from .sized_member import sized_member
from .structureview import StructureView

bitfield_member.__module__ = "plum.structure"
dimmed_member.__module__ = "plum.structure"
member.__module__ = "plum.structure"
StructureMeta.__module__ = "plum.structure"
sized_member.__module__ = "plum.structure"


class Structure(list, Data, metaclass=StructureMeta):

    """Structured data store type."""

    # filled in by metaclass
    __ignore_flags__: Tuple[bool, ...] = ()
    __members__: Tuple[Member, ...] = ()
    __nbytes__: Union[None, int] = 0
    __offsets__: Union[None, Tuple[int, ...]] = ()
    __byteorder__: str = "little"
    __implementation__: Optional[str] = None
    __format_name__: str
    __hint__: str

    # Metaclass generates the following methods for each subclass:
    #   __init__
    #   member getter/setters
    #   __eq__
    #   __ne__
    #   __pack__
    #   __pack_and_dump__
    #   __repr__
    #   __unpack__
    #   __unpack_and_dump__

    def asdict(self) -> Dict[str, Any]:
        """Return structure members in dictionary form.

        :returns: structure members

        """
        return {
            name: getattr(self, name)
            for name in (member.name for member in self.__members__)
        }

    @classmethod
    def _make_structure_from_dict(cls, members: Dict[str, Any]) -> "Structure":
        return cls(**members)

    def __setattr__(self, name, value):
        # get the attribute to raise an exception if invalid name
        getattr(self, name)
        object.__setattr__(self, name, value)

    @classmethod
    def __view__(cls, buffer, offset: int = 0):
        """Create plum view of bytes buffer.

        :param buffer: bytes buffer
        :type buffer: bytes-like (e.g. bytes, bytearray, memoryview)
        :param offset: byte offset

        """
        if not cls.__nbytes__:
            raise TypeError(
                f"cannot create view for structure {cls.__name__!r} "
                "with variable size"
            )

        return StructureView(cls, buffer, offset)


__all__ = [
    "bitfield_member",
    "dimmed_member",
    "member",
    "StructureMeta",
    "sized_member",
    "Structure",
]



# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Copyright 2021 Daniel Mark Gass, see __about__.py for license information.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
"""Test structure data store transform properties."""

# pylint doesn't seem happy with properties in metaclass
# pylint: disable=comparison-with-callable

import pytest

from plum.array import ArrayX
from plum.bigendian import uint8
from plum.exceptions import SizeError
from plum.structure import Structure, member


class TestDefault:
    """Test with as many left to default as possible."""

    class Struct(Structure):
        m0 = member(fmt=uint8)

    def test_byteorder(self):
        assert self.Struct.byteorder == "little"

    def test_nbytes(self):
        assert self.Struct.nbytes == 1
        assert self.Struct(m0=0).nbytes == 1


class TestKeyword:

    """Test explicitly defined with keyword argument."""

    class Struct(Structure, byteorder="big"):
        m0 = member(fmt=ArrayX(name="array", fmt=uint8))

    def test_byteorder(self):
        assert self.Struct.byteorder == "big"

    def test_nbytes(self):
        with pytest.raises(SizeError):
            self.Struct.nbytes  # pylint: disable=pointless-statement
        assert self.Struct(m0=[0]).nbytes == 1



# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Copyright 2022 Daniel Mark Gass, see __about__.py for license information.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
"""Test OptionalX transform properties."""

import pytest
from plum.exceptions import SizeError
from plum.littleendian import uint8
from plum.optional import OptionalX
from plum.structure import Structure, member


class Sample(Structure):

    member1: int = member(fmt=uint8)


class TestDefaultTransform:

    optional_x = OptionalX(fmt=uint8)

    def test_hint(self):
        assert self.optional_x.__hint__ == "Optional[int]"

    def test_name(self):
        assert self.optional_x.name == "Optional[uint8]"

    def test_fmt(self):
        assert self.optional_x.fmt is uint8

    def test_nbytes(self):
        with pytest.raises(SizeError):
            self.optional_x.nbytes  # pylint: disable=pointless-statement


class TestDefaultTransformClass:

    optional_x = OptionalX(fmt=Sample)

    def test_hint(self):
        assert self.optional_x.__hint__ == "Optional[Sample]"

    def test_name(self):
        assert self.optional_x.name == "Optional[Sample (Structure)]"

    def test_fmt(self):
        assert self.optional_x.fmt is Sample

    def test_nbytes(self):
        with pytest.raises(SizeError):
            self.optional_x.nbytes  # pylint: disable=pointless-statement


class TestPosition:

    optional_x = OptionalX(Sample, "name")

    def test_hint(self):
        assert self.optional_x.__hint__ == "Optional[Sample]"

    def test_name(self):
        assert self.optional_x.name == "name"

    def test_fmt(self):
        assert self.optional_x.fmt is Sample

    def test_nbytes(self):
        with pytest.raises(SizeError):
            self.optional_x.nbytes  # pylint: disable=pointless-statement


class TestKeyword(TestPosition):

    optional_x = OptionalX(fmt=Sample, name="name")



# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Copyright 2021 Daniel Mark Gass, see __about__.py for license information.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
"""Test common exceptions using BitFields data store transform."""

import pytest
from baseline import Baseline

from plum.bitfields import BitFields, bitfield
from plum.conformance import wrap_message

from sample_bitfields import MyBits


class TestExceptions:
    def test_attribute_fat_finger(self):
        """Test invalid attribute set."""
        expected_message = Baseline(
            """
            'MyBits' has no attribute 'junk'
            """
        )

        mybits = MyBits.from_int(0)

        with pytest.raises(AttributeError) as trap:
            # pylint: disable=attribute-defined-outside-init
            mybits.junk = 1

        assert wrap_message(trap.value) == expected_message

    def test_already_in_use(self):
        """Test bit field definition may only be used in one BitsField class."""

        expected = Baseline(
            """
            invalid bit field 'xyz' definition, BitField() instance can not be
            shared between bit fields classes
            """
        )

        bitfield_ = bitfield(typ=int, lsb=0, size=1)

        class FirstUse(BitFields, nbytes=1):  # pylint: disable=unused-variable

            """Sample bit fields."""

            xyz = bitfield_

        with pytest.raises(TypeError) as trap:

            class SecondUse(BitFields, nbytes=1):  # pylint: disable=unused-variable
                """Sample bit fields."""

                xyz = bitfield_

        assert wrap_message(trap.value) == expected

    def test_from_int_too_small(self):
        expected_message = Baseline(
            """
            MyBits.from_int() requires 0 <= value <= 65535
            """
        )

        with pytest.raises(ValueError) as trap:
            MyBits.from_int(-1)

        assert wrap_message(trap.value) == expected_message

    def test_from_int_too_big(self):
        expected_message = Baseline(
            """
            MyBits.from_int() requires 0 <= value <= 65535
            """
        )

        with pytest.raises(ValueError) as trap:
            MyBits.from_int(0x10000)

        assert wrap_message(trap.value) == expected_message



# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Copyright 2021 Daniel Mark Gass, see __about__.py for license information.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
"""Test pack() utility function."""

import pytest
from baseline import Baseline

from plum.conformance import wrap_message
from plum.exceptions import PackError
from plum.littleendian import uint8, uint16
from plum.utilities import pack, pack_and_dump


class TestPackHappyPath:

    """Test normal operations packing multiple items in one call."""

    def test_happy_path_single(self):
        """Test packing with single pos instance args."""
        exp_buffer = b"\x02\x01"
        exp_dump = Baseline(
            """
            +--------+-------+-------+--------+
            | Offset | Value | Bytes | Format |
            +--------+-------+-------+--------+
            | 0      | 258   | 02 01 | uint16 |
            +--------+-------+-------+--------+
            """
        )

        fmt = uint16

        assert pack(0x0102, fmt) == exp_buffer

        buffer, dump = pack_and_dump(0x0102, fmt)
        assert (buffer, str(dump)) == (exp_buffer, exp_dump)

    def test_happy_path_items(self):
        """Test packing with pos instance args."""
        exp_buffer = b"\x03\x02\x01"
        exp_dump = Baseline(
            """
            +--------+--------+-------+-------+--------+
            | Offset | Access | Value | Bytes | Format |
            +--------+--------+-------+-------+--------+
            | 0      | [0]    | 3     | 03    | uint8  |
            | 1      | [1]    | 258   | 02 01 | uint16 |
            +--------+--------+-------+-------+--------+
            """
        )

        fmt = (uint8, uint16)

        assert pack((3, 0x0102), fmt) == exp_buffer

        buffer, dump = pack_and_dump((3, 0x0102), fmt)
        assert (buffer, str(dump)) == (exp_buffer, exp_dump)

    def test_happy_path_kwitems(self):
        """Test packing with kw instance args."""
        exp_buffer = b"\x00\x02\x01"
        exp_dump = Baseline(
            """
            +--------+--------+-------+-------+--------+
            | Offset | Access | Value | Bytes | Format |
            +--------+--------+-------+-------+--------+
            | 0      | ['m1'] | 0     | 00    | uint8  |
            | 1      | ['m2'] | 258   | 02 01 | uint16 |
            +--------+--------+-------+-------+--------+
            """
        )

        fmt = {"m1": uint8, "m2": uint16}

        assert pack(dict(m1=0, m2=0x0102), fmt) == exp_buffer

        buffer, dump = pack_and_dump(dict(m1=0, m2=0x0102), fmt)
        assert (buffer, str(dump)) == (exp_buffer, exp_dump)

    def test_happy_path_nested(self):
        """Test packing with both named and unnamed values (nested)."""
        exp_buffer = b"\x00\x02\x01"
        exp_dump = Baseline(
            """
            +--------+--------+-------+-------+--------+
            | Offset | Access | Value | Bytes | Format |
            +--------+--------+-------+-------+--------+
            |        | ['m1'] |       |       |        |
            | 0      |   [0]  | 0     | 00    | uint8  |
            | 1      |   [1]  | 258   | 02 01 | uint16 |
            +--------+--------+-------+-------+--------+
            """
        )

        fmt = {"m1": (uint8, uint16)}

        assert pack(dict(m1=(0, 0x0102)), fmt) == exp_buffer

        buffer, dump = pack_and_dump(dict(m1=(0, 0x0102)), fmt)
        assert (buffer, str(dump)) == (exp_buffer, exp_dump)

    def test_empty_dict_fmt(self):
        """Test packing with both named and unnamed values (nested)."""
        exp_buffer = b""
        exp_dump = Baseline(
            """
            +--------+-------+-------+--------+
            | Offset | Value | Bytes | Format |
            +--------+-------+-------+--------+
            |        |       |       |        |
            +--------+-------+-------+--------+
            """
        )

        fmt = {}

        assert pack({}, fmt) == exp_buffer

        buffer, dump = pack_and_dump({}, fmt)
        assert (buffer, str(dump)) == (exp_buffer, exp_dump)


class TestPlumView:

    """Test pack() operates with plum view types."""

    def test_view_as_pos_argument(self):
        """Test packing with view as positional argument."""
        exp_buffer = b"\x00\x02\x01"
        exp_dump = Baseline(
            """
            +--------+--------+-------+-------+--------+
            | Offset | Access | Value | Bytes | Format |
            +--------+--------+-------+-------+--------+
            | 0      | [0]    | 0     | 00    | uint8  |
            | 1      | [1]    | 258   | 02 01 | uint16 |
            +--------+--------+-------+-------+--------+
            """
        )

        shared_memory = b"\x02\x01\x00"
        uint16_view = uint16.view(shared_memory, offset=0)
        uint8_view = uint8.view(shared_memory, offset=2)

        fmt = (uint8, uint16)

        assert pack((uint8_view, uint16_view), fmt) == exp_buffer

        buffer, dump = pack_and_dump((uint8_view, uint16_view), fmt)
        assert (buffer, str(dump)) == (exp_buffer, exp_dump)

    def test_view_as_kw_argument(self):
        """Test packing with view as keyword argument."""

        exp_buffer = b"\x00\x02\x01"
        exp_dump = Baseline(
            """
            +--------+--------+-------+-------+--------+
            | Offset | Access | Value | Bytes | Format |
            +--------+--------+-------+-------+--------+
            | 0      | ['m1'] | 0     | 00    | uint8  |
            | 1      | ['m2'] | 258   | 02 01 | uint16 |
            +--------+--------+-------+-------+--------+
            """
        )

        shared_memory = b"\x02\x01\x00"
        uint16_view = uint16.view(shared_memory, offset=0)
        uint8_view = uint8.view(shared_memory, offset=2)

        fmt = {"m1": uint8, "m2": uint16}

        assert pack(dict(m1=uint8_view, m2=uint16_view), fmt) == exp_buffer

        buffer, dump = pack_and_dump(dict(m1=uint8_view, m2=uint16_view), fmt)
        assert (buffer, str(dump)) == (exp_buffer, exp_dump)


class TestExceptions:

    """Test pack exception corner cases."""

    def test_bad_single_type(self):
        """Test packing exception with format not a valid choice."""
        expected = Baseline(
            """
            +--------+-------+-------+---------------+
            | Offset | Value | Bytes | Format        |
            +--------+-------+-------+---------------+
            |        | 0     |       | 258 (invalid) |
            +--------+-------+-------+---------------+

            TypeError occurred during pack operation:

            bad item format, must be a packable data type/transform (or a dict,
            list, or tuple of them)
            """
        )

        with pytest.raises(PackError) as trap:
            pack((0, 0x0102))  # bad fmt
        assert wrap_message(trap.value) == expected
        assert isinstance(trap.value.__context__, TypeError)

        with pytest.raises(PackError) as trap:
            pack_and_dump((0, 0x0102))  # bad fmt
        assert wrap_message(trap.value) == expected
        assert isinstance(trap.value.__context__, TypeError)

    def test_bad_tuple_type(self):
        """Test packing exception with invalid type in format tuple."""
        expected = Baseline(
            """
            +--------+--------+-------+-------+-------------+
            | Offset | Access | Value | Bytes | Format      |
            +--------+--------+-------+-------+-------------+
            | 0      | [0]    | 0     | 00    | uint8       |
            |        | [1]    | 258   |       | 0 (invalid) |
            +--------+--------+-------+-------+-------------+

            TypeError occurred during pack operation:

            bad item format, must be a packable data type/transform (or a dict,
            list, or tuple of them)
            """
        )

        with pytest.raises(PackError) as trap:
            pack((0, 0x0102), (uint8, 0))
        assert wrap_message(trap.value) == expected
        assert isinstance(trap.value.__context__, TypeError)

        with pytest.raises(PackError) as trap:
            pack_and_dump((0, 0x0102), (uint8, 0))
        assert wrap_message(trap.value) == expected
        assert isinstance(trap.value.__context__, TypeError)

    def test_bad_dict_type(self):
        """Test packing exception with invalid type in format tuple."""
        expected = Baseline(
            """
            +--------+--------+-------+-------+-------------+
            | Offset | Access | Value | Bytes | Format      |
            +--------+--------+-------+-------+-------------+
            | 0      | ['m1'] | 0     | 00    | uint8       |
            |        | ['m2'] | 258   |       | 0 (invalid) |
            +--------+--------+-------+-------+-------------+

            TypeError occurred during pack operation:

            bad item format, must be a packable data type/transform (or a dict,
            list, or tuple of them)
            """
        )

        fmt = {"m1": uint8, "m2": 0}

        with pytest.raises(PackError) as trap:
            pack(dict(m1=0, m2=0x0102), fmt)
        assert wrap_message(trap.value) == expected
        assert isinstance(trap.value.__context__, TypeError)

        with pytest.raises(PackError) as trap:
            pack_and_dump(dict(m1=0, m2=0x0102), fmt)
        assert wrap_message(trap.value) == expected
        assert isinstance(trap.value.__context__, TypeError)

    def test_args_shortage(self):
        """Test packing exception with not enough positional args."""
        expected = Baseline(
            """
            +--------+--------+-----------+-------+--------+
            | Offset | Access | Value     | Bytes | Format |
            +--------+--------+-----------+-------+--------+
            | 0      | [0]    | 0         | 00    | uint8  |
            |        | [1]    | (missing) |       | uint16 |
            +--------+--------+-----------+-------+--------+

            TypeError occurred during pack operation:

            1 value given, expected 2
            """
        )

        fmt = (uint8, uint16)

        with pytest.raises(PackError) as trap:
            pack([0], fmt)
        assert wrap_message(trap.value) == expected
        assert isinstance(trap.value.__context__, TypeError)

        with pytest.raises(PackError) as trap:
            pack_and_dump([0], fmt)
        assert wrap_message(trap.value) == expected
        assert isinstance(trap.value.__context__, TypeError)

    def test_args_excess(self):
        """Test packing exception with too many positional args."""
        expected = Baseline(
            """
            +--------+--------+-------+-------+--------------+
            | Offset | Access | Value | Bytes | Format       |
            +--------+--------+-------+-------+--------------+
            | 0      | [0]    | 0     | 00    | uint8        |
            | 1      | [1]    | 1     | 01 00 | uint16       |
            +--------+--------+-------+-------+--------------+
            |        | [2]    | 2     |       | (unexpected) |
            +--------+--------+-------+-------+--------------+

            TypeError occurred during pack operation:

            3 values given, expected 2
            """
        )

        fmt = (uint8, uint16)

        with pytest.raises(PackError) as trap:
            pack((0, 1, 2), fmt)
        assert wrap_message(trap.value) == expected
        assert isinstance(trap.value.__context__, TypeError)

        with pytest.raises(PackError) as trap:
            pack_and_dump((0, 1, 2), fmt)
        assert wrap_message(trap.value) == expected
        assert isinstance(trap.value.__context__, TypeError)

    def test_kwargs_shortage(self):
        """Test packing exception with not enough keyword args."""
        expected = Baseline(
            """
            +--------+--------+-----------+-------+--------+
            | Offset | Access | Value     | Bytes | Format |
            +--------+--------+-----------+-------+--------+
            | 0      | ['m1'] | 0         | 00    | uint8  |
            |        | ['m2'] | (missing) |       | uint16 |
            +--------+--------+-----------+-------+--------+

            TypeError occurred during pack operation:

            missing value: 'm2'
            """
        )

        fmt = {"m1": uint8, "m2": uint16}

        with pytest.raises(PackError) as trap:
            pack(dict(m1=0), fmt)
        assert wrap_message(trap.value) == expected
        assert isinstance(trap.value.__context__, TypeError)

        with pytest.raises(PackError) as trap:
            pack_and_dump(dict(m1=0), fmt)
        assert wrap_message(trap.value) == expected
        assert isinstance(trap.value.__context__, TypeError)

    def test_kwargs_excess(self):
        """Test packing exception with too many keyword args."""
        expected = Baseline(
            """
            +--------+--------+-------+-------+--------------+
            | Offset | Access | Value | Bytes | Format       |
            +--------+--------+-------+-------+--------------+
            | 0      | ['m1'] | 0     | 00    | uint8        |
            | 1      | ['m2'] | 1     | 01 00 | uint16       |
            +--------+--------+-------+-------+--------------+
            |        | ['m3'] | 2     |       | (unexpected) |
            +--------+--------+-------+-------+--------------+

            TypeError occurred during pack operation:

            unexpected value: 'm3'
            """
        )

        fmt = {"m1": uint8, "m2": uint16}

        with pytest.raises(PackError) as trap:
            pack(dict(m1=0, m2=1, m3=2), fmt)
        assert wrap_message(trap.value) == expected
        assert isinstance(trap.value.__context__, TypeError)

        with pytest.raises(PackError) as trap:
            pack_and_dump(dict(m1=0, m2=1, m3=2), fmt)
        assert wrap_message(trap.value) == expected
        assert isinstance(trap.value.__context__, TypeError)



# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Copyright 2021 Daniel Mark Gass, see __about__.py for license information.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
"""Test Structure representation."""

# pylint: disable=unexpected-keyword-arg

from typing import Type

from baseline import Baseline

from plum.bitfields import BitFields, bitfield
from plum.enum import EnumX

from sample_bitfields import Register


class Base:

    """Test case base class."""

    Sample: Type[BitFields]
    sample: BitFields
    implementation_baseline: Baseline
    repr_baseline: Baseline

    @staticmethod
    def iter_lines(code):
        lines = code.split("\n")

        active = False

        for line in lines:
            if "__repr__" in line:
                active = True

            if active:
                yield line

            if not line.strip():
                active = False

    def test_generated_code(self):
        compare_implementation = "\n".join(self.iter_lines(self.Sample.implementation))
        assert compare_implementation == self.implementation_baseline

    def test_repr(self):
        assert repr(self.sample) == self.repr_baseline


class TestVanilla(Base):

    """Test default structure representation (one enum, one int)."""

    class Sample(BitFields):

        """Sample bit fields data store transform type."""

        m0: int = bitfield(size=1)
        m1: int = bitfield(size=1, typ=Register)

    sample = Sample(m0=0, m1=Register.SP)

    repr_baseline = Baseline(
        """
        Sample(m0=0, m1=Register.SP)
        """
    )

    implementation_baseline = Baseline(
        """
        def __repr__(self) -> str:
            try:
                return f"{type(self).__name__}(m0={self.m0!r}, m1={repr(self.m1).split(':')[0].lstrip('<')})"
            except Exception:
                return f"{type(self).__name__}()"

        """
    )

    def test_repr(self):
        actual_repr = repr(self.sample)  # pylint: disable=no-member

        assert actual_repr == self.repr_baseline  # pylint: disable=no-member


class TestTolerantEnum(Base):

    """Test structure representation with invalid enum member."""

    class Sample(BitFields):

        """Sample bit fields data store transform type."""

        m0: int = bitfield(size=4)
        m1: int = bitfield(size=4, typ=EnumX(enum=Register, strict=False))

    sample = Sample(m0=0, m1=15)

    repr_baseline = Baseline(
        """
        Sample(m0=0, m1=15)
        """
    )

    implementation_baseline = Baseline(
        """
        def __repr__(self) -> str:
            try:
                return f"{type(self).__name__}(m0={self.m0!r}, m1={repr(self.m1).split(':')[0].lstrip('<')})"
            except Exception:
                return f"{type(self).__name__}()"

        """
    )

    def test_repr(self):
        actual_repr = repr(self.sample)  # pylint: disable=no-member

        assert actual_repr == self.repr_baseline  # pylint: disable=no-member


class TestBlank(Base):

    """Test structure representation when one member specified as nothing."""

    class Sample(BitFields):

        """Sample bit fields data store transform type."""

        m0: int = bitfield(size=1, argrepr="")
        m1: int = bitfield(size=1)

    sample = Sample(m0=0, m1=1)

    repr_baseline = Baseline(
        """
        Sample(m1=1)
        """
    )

    implementation_baseline = Baseline(
        """
        def __repr__(self) -> str:
            try:
                return f"{type(self).__name__}(m1={self.m1!r})"
            except Exception:
                return f"{type(self).__name__}()"

        """
    )


class TestCustom(Base):

    """Test structure representation when one member has custom format."""

    class Sample(BitFields):

        """Sample bit fields data store transform type."""

        m0: int = bitfield(size=1, argrepr="product={self.m0 * self.m1}")
        m1: int = bitfield(size=1)

    sample = Sample(m0=0, m1=1)

    repr_baseline = Baseline(
        """
        Sample(product=0, m1=1)
        """
    )

    implementation_baseline = Baseline(
        """
        def __repr__(self) -> str:
            try:
                return f"{type(self).__name__}(product={self.m0 * self.m1}, m1={self.m1!r})"
            except Exception:
                return f"{type(self).__name__}()"

        """
    )


class TestOverride:

    """Test that custom __repr__ method is not overridden."""

    class Sample(BitFields):

        """Sample bit fields data store transform type."""

        m0: int = bitfield(size=1, argrepr="product={_m0 * _m1}")
        m1: int = bitfield(size=1)

        def __repr__(self):
            return "CustomRepresentation"

    def test_repr(self):
        assert repr(self.Sample(m0=0, m1=1)) == "CustomRepresentation"



# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Copyright 2022 Daniel Mark Gass, see __about__.py for license information.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
"""Bit fields data store type."""

# pylint: disable=too-many-arguments

from enum import EnumMeta
from typing import Any, Optional

from ..enum import EnumX
from ..flag import FlagX


class BitField(property):

    """Bit field definition.

    :param doc: one line description
    :param size: size in bits
    :param typ: bit field type, default uses type annotation
    :param lsb: bit offset of least significant bit
    :param default: initial value when unspecified
    :param signed: interpret as signed integer
    :param ignore: do not include field in comparisons
    :param readonly: block setting member attribute
    :param argrepr: format to represent member argument in structure repr

    """

    # pylint: disable=too-many-instance-attributes

    def __init__(
        self,
        doc: str,
        size: int,
        typ: type,
        lsb: Optional[int],
        default: Optional[int],
        signed: bool,
        ignore: bool,
        readonly: bool,
        argrepr: Optional[str],
    ):
        property.__init__(self)

        if lsb is not None:
            lsb = int(lsb)
            if lsb < 0:
                raise TypeError(
                    "bit field position must be greater than or equal to zero"
                )

        size = int(size)
        signed = bool(signed)

        if signed:
            if size < 2:
                raise ValueError("'size' must be 2 or greater for signed bit field")
            signbit = 1 << (size - 1)
        else:
            if size < 1:
                raise ValueError("'size' must be 1 or greater for unsigned bit field")
            signbit = 0

        if signbit:
            minvalue = -(1 << (size - 1))
            maxvalue = -(1 + minvalue)
        else:
            minvalue = 0
            maxvalue = (1 << size) - 1

        if default is not None:
            default = int(default)
            if not minvalue <= default <= maxvalue:
                raise ValueError(
                    f"bit field requires {minvalue} <= default <= {maxvalue}"
                )

        self._doc = doc
        self._default = default
        self._ignore = ignore
        self._mask = (1 << size) - 1
        self._minvalue = minvalue
        self._maxvalue = maxvalue
        self._name: Optional[str] = None  # filled in later by BitFieldsMeta
        self._lsb = lsb
        self._signbit = signbit
        self._size = size
        self._make_type = None
        self._readonly = readonly
        self._argrepr = argrepr
        self._set_typ(typ)

    @property
    def argrepr(self) -> str:
        """Format string to represent bit field argument in bit fields repr."""
        if isinstance(self._argrepr, str):
            repr_format = self._argrepr

        elif isinstance(self._typ, (EnumMeta, EnumX)):
            repr_format = (
                f"{self.name}={{repr(self.{self.name}).split(':')[0].lstrip('<')}}"
            )

        else:
            repr_format = f"{self.name}={{self.{self.name}!r}}"

        return repr_format

    @property
    def default(self) -> Optional[int]:
        """Initializer default value for this bitfield."""
        return self._default

    @property
    def doc(self) -> str:
        """One line bit field description."""
        return self._doc

    @property
    def ignore(self) -> bool:
        """Ignore when compared against bitfields of same type."""
        return self._ignore

    @property
    def lsb(self) -> int:
        """Bit offset of least significant bit."""
        if self._lsb is None:  # pragma: no cover, for mypy purposes
            raise RuntimeError("internal error")
        return self._lsb

    @property
    def mask(self) -> int:
        """Bit zero aligned bit mask covering the width of the bit field."""
        return self._mask

    @property
    def minvalue(self) -> int:
        """Minimum allowable bit field value."""
        return self._minvalue

    @property
    def maxvalue(self) -> int:
        """Maximum allowable bit field value."""
        return self._maxvalue

    @property
    def name(self) -> str:
        """Bit field name."""
        if self._name is None:  # pragma: no cover, for mypy purposes
            raise RuntimeError("internal error")
        return self._name

    @property
    def readonly(self) -> bool:
        """Block setting bit field attribute."""
        return self._readonly

    @property
    def signbit(self) -> int:
        """Mask when anded with bit field value reveals if value is negative."""
        return self._signbit

    @property
    def signed(self) -> bool:
        """Interpret bit field as signed integer."""
        return bool(self._signbit)

    @property
    def size(self) -> int:
        """Bit field size in bits."""
        return self._size

    @property
    def type_hint(self) -> type:
        """Type hint expression."""
        typ = self._typ

        if isinstance(typ, EnumX):
            typ = typ.enum

        return typ

    @property
    def typ(self) -> type:
        """Bit field type.

        :returns: bit field type
        :rtype: type

        """
        return self._typ

    def _set_typ(self, typ: type):
        is_int_subclass = isinstance(typ, type) and issubclass(typ, int)

        if isinstance(typ, (FlagX, EnumX)) or is_int_subclass:
            pass

        elif hasattr(typ, "nested"):  # must be BitFields
            if not getattr(typ, "nested"):
                name = f"{self.name!r} " if self._name else ""
                raise TypeError(
                    f"bit field typ {name}must be declared as nested (e.g. "
                    f"'class {typ.__name__}(BitFields, nested=True):')"
                )

        else:
            raise TypeError("bit field type must be int-like")

        self._typ = typ

    def finish_initialization(self, name: str) -> None:
        """Complete instance initialization.

        :param name: member name

        """
        if self._name is None:
            self._name = name
        else:
            raise TypeError(
                f"invalid bit field {name!r} definition, "
                f"{type(self).__name__}() instance can not be shared "
                f"between bit fields classes"
            )

    def __repr__(self):
        return f"BitField(name={self._name!r})"

    def getter(self, fget):
        property.__init__(self, fget, self.fset, self.fdel)
        return self

    def setter(self, fset):
        if self.readonly:
            raise TypeError("'setter' not allowed on read-only bit field properties")

        property.__init__(self, self.fget, fset, self.fdel)
        return self

    def deleter(self, fdel):
        raise TypeError("bit field properties do not support 'deleter'")


def bitfield(
    doc: str = "",
    *,
    size: int,
    typ: type = int,
    lsb: Optional[int] = None,
    default: Optional[int] = None,
    signed: bool = False,
    ignore: bool = False,
    readonly: bool = False,
    argrepr: Optional[str] = None,
) -> Any:
    """Define bit field.

    :param doc: one line description
    :param size: size in bits
    :param typ: bit field type
    :param lsb: bit offset of least significant bit
    :param default: initial value when unspecified
    :param signed: interpret as signed integer
    :param ignore: do not include field in comparisons
    :param readonly: block setting member attribute
    :param argrepr: format to represent member argument in structure repr

    """
    return BitField(doc, size, typ, lsb, default, signed, ignore, readonly, argrepr)



# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Copyright 2021 Daniel Mark Gass, see __about__.py for license information.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
"""Test "strict" feature (where strict=False)."""

# pylint: disable=unidiomatic-typecheck

from baseline import Baseline

from plum.enum import EnumX

from sample_enum import Register

register = EnumX(Register, 1, "big", signed=False, strict=False, name="register")


class TestPack:

    """Test pack variants with invalid value."""

    expected_dump = Baseline(
        """
        +--------+-------+-------+----------+
        | Offset | Value | Bytes | Format   |
        +--------+-------+-------+----------+
        | 0      | 4     | 04    | register |
        +--------+-------+-------+----------+
        """
    )

    def test_pack(self):
        assert register.pack(4) == b"\x04"

    def test_pack_and_dump(self):
        buffer, dump = register.pack_and_dump(4)
        assert buffer == b"\x04"
        assert str(dump) == self.expected_dump


class TestUnpack:

    """Test unpack variants with invalid value."""

    expected_dump = Baseline(
        """
        +--------+-------+-------+----------+
        | Offset | Value | Bytes | Format   |
        +--------+-------+-------+----------+
        | 0      | 4     | 04    | register |
        +--------+-------+-------+----------+
        """
    )

    def test_unpack(self):
        value = register.unpack(b"\x04")
        assert value == 4
        assert type(value) is int

    def test_unpack_and_dump(self):
        value, dump = register.unpack_and_dump(b"\x04")
        assert str(dump) == self.expected_dump
        assert value == 4
        assert type(value) is int


class TestTransformCall:

    """Test transform call."""

    def test_call(self):
        assert register(4) == 4
        assert type(register(4)) is int



# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Copyright 2021 Daniel Mark Gass, see __about__.py for license information.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
"""Test items transform with fmt=None and nested structure."""

from baseline import Baseline
from plum.items import items
from plum.littleendian import uint8


class Test:

    expected_dump = Baseline(
        """
        +--------+----------+-------+-------+------------+
        | Offset | Access   | Value | Bytes | Format     |
        +--------+----------+-------+-------+------------+
        |        |          |       |       | items:dict |
        |        | ['mbr1'] |       |       | list       |
        | 0      |   [0]    | 1     | 01    | uint8      |
        | 1      |   [1]    | 2     | 02    | uint8      |
        | 2      | ['mbr2'] | 3     | 03    | uint8      |
        +--------+----------+-------+-------+------------+
        """
    )

    expected_bytes = bytes.fromhex("010203")

    sample = dict(mbr1=[(1, uint8), (2, uint8)], mbr2=(3, uint8))

    def test_pack(self):
        assert items.pack(self.sample) == self.expected_bytes

    def test_pack_and_dump(self):
        memory, dump = items.pack_and_dump(self.sample)

        assert memory == self.expected_bytes
        assert str(dump) == self.expected_dump



# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Copyright 2021 Daniel Mark Gass, see __about__.py for license information.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
"""Structure methods generator."""

from typing import Any, Dict, Generator, List, Tuple, ValuesView

from .member import Member


class CodeMaker:

    """Structure method code generator."""

    members: tuple
    names_tuple: str

    def __init__(
        self,
        structure_name: str,
        members: Tuple[Member, ...],
        new_members: ValuesView[Member],
    ) -> None:
        names = [member.temp_store.local_name for member in members]

        names_tuple = ", ".join(name for name in names if name)

        if "," not in names_tuple:
            names_tuple += ", "

        self.structure_name = structure_name
        self.members = members
        self.names_tuple = names_tuple
        self.new_members = new_members

    def iter_init_lines(self) -> Generator[str, None, None]:
        """Generate __init__ method lines."""
        indent = "    "

        member_stores = [member.temp_store for member in self.members]
        parameters = [store.init_parameter for store in member_stores]
        args = ", ".join(parameters)

        yield f"def __init__(self, *, {args}) -> None:"

        def lines_from_store(store_name: str):
            nonlocal member_stores, indent
            for member_store in member_stores:
                lines = getattr(member_store, store_name)
                if lines:
                    for line in lines:
                        yield (indent + line).rstrip()
                    yield ""

        yield from lines_from_store("init_pre_begin")

        yield from lines_from_store("init_begin")

        yield indent + f'self[:] = ({self.names_tuple[2:].replace(", m_", ", ")})'
        yield ""

        yield from lines_from_store("init_end")

    def iter_pack_lines(self) -> Generator[str, None, None]:
        """Structure __pack__ method generator."""
        indent = "    "

        yield "@classmethod"
        yield "def __pack__(cls, value, pieces: List[bytes], dump: Optional[Record] = None) -> None:"

        yield indent + "if isinstance(value, dict):"
        yield indent + "    value = cls._make_structure_from_dict(value)"
        yield ""

        yield indent + f"({self.names_tuple}) = value"
        yield ""

        yield indent + "if dump is None:"

        for member in self.members:
            lines = getattr(member.temp_store, "pack")
            for line in lines:
                yield indent + indent + line
            yield ""

        yield indent + "else:"

        for member in self.members:
            lines = getattr(member.temp_store, "pack_and_dump")
            for line in lines:
                yield indent + indent + line
            yield ""

    def iter_unpack_lines(self) -> Generator[str, None, None]:
        """Structure __unpack__ method generator."""
        indent = "    "

        factory_present = any(
            m.fmt_is_a_factory and m.fmt_arg is None for m in self.members
        )

        yield "@classmethod"
        yield f'def __unpack__(cls, buffer: bytes, offset: int, dump: Optional[Record] = None) -> Tuple["{self.structure_name}", int]:'
        yield indent + "structure = list.__new__(cls)"
        if factory_present:
            yield indent + "add_member = structure.append"
        yield ""

        yield indent + "if dump is None:"

        for member in self.members:
            for line in member.temp_store.unpack:
                yield indent + indent + line
            if factory_present:
                yield indent + indent + f"add_member(m_{member.name})"
            for line in member.temp_store.unpack_checks:
                yield indent + indent + line
            yield ""

        yield indent + "else:"

        for member in self.members:
            for line in member.temp_store.unpack_and_dump:
                yield indent + indent + line
            if factory_present:
                yield indent + indent + f"add_member(m_{member.name})"
            for line in member.temp_store.unpack_checks:
                yield indent + indent + line
            yield ""

        if not factory_present:
            yield indent + f"structure[:] = ({self.names_tuple})"
            yield ""

        yield indent + "return structure, offset"
        yield ""

    def iter_compare_unfold(
        self, source: str, compute_members: List[str]
    ) -> Generator[str, None, None]:
        """Generate lines to get local variable for each member."""
        indent = " " * 8
        prefix = f"_{source[0]}_"
        left_side = ", ".join(prefix + m.name for m in self.members)
        compute_members = [prefix + name for name in compute_members]

        yield indent + f"{left_side} = {source}"

        if len(compute_members) == 1:
            yield indent + f"if {compute_members[0]} is None:"
            yield indent + f"    {left_side} = self.unpack({source}.ipack())"
        elif compute_members:
            yield indent + f"if None in ({', '.join(compute_members)}):"
            yield indent + f"    {left_side} = self.unpack({source}.ipack())"

    def iter_compare(self, name: str) -> Generator[str, None, None]:
        """Structure __eq__ and __ne__ method generator."""
        compute_members = [m.name for m in self.members if m.compute and not m.ignore]
        names = [m.name for m in self.members if not m.ignore]

        yield f"def __{name}__(self, other: Any) -> bool:"
        yield "    if isinstance(other, dict):"
        yield "        other = self._make_structure_from_dict(other)"

        if (
            not names
            or any(member.ignore for member in self.members)
            or compute_members
        ):
            yield "    elif isinstance(other, type(self)):"
            if names:
                left = ", ".join("_s_" + n for n in names)
                right = ", ".join("_o_" + n for n in names)
                if len(names) > 1:
                    left = f"({left})"
                    right = f"({right})"
                operator = "==" if name == "eq" else "!="
                yield from self.iter_compare_unfold("self", compute_members)
                yield from self.iter_compare_unfold("other", compute_members)
                yield f"        return {left} {operator} {right}"
            else:
                yield f"        return {name == 'eq'}"

            yield "    else:"
            yield f"        return list.__{name}__(self, other)"
        else:
            yield f"    return list.__{name}__(self, other)"

        yield ""

    def iter_repr(self) -> Generator[str, None, None]:
        """Structure __repr__ method generator."""
        member_names = ["_" + m.name for m in self.members]
        if len(member_names) == 1:
            member_names.append("")

        arg_reprs = ", ".join(m.argrepr for m in self.members if m.argrepr)
        class_name = "{type(self).__name__}"

        yield "def __repr__(self) -> str:"
        yield "    try:"
        yield f'        return f"{class_name}({arg_reprs})"'
        yield "    except Exception:"
        yield f'        return f"{class_name}()"'
        yield ""

    def iter_lines(self, namespace: Dict[str, Any]) -> Generator[str, None, None]:
        """Generate structure class method code lines."""
        if "__init__" not in namespace:
            yield from self.iter_init_lines()

        for member in self.new_members:
            yield from member.iter_getter_lines()
            yield from member.iter_setter_lines()

        if "__pack__" not in namespace:
            yield from self.iter_pack_lines()

        if "__unpack__" not in namespace:
            yield from self.iter_unpack_lines()

        if "__eq__" not in namespace:
            yield from self.iter_compare("eq")

        if "__ne__" not in namespace:
            yield from self.iter_compare("ne")

        if "__repr__" not in namespace:
            yield from self.iter_repr()
